import traceback
from collections import defaultdict

from sqlalchemy import null, or_, select
from sqlalchemy.orm import joinedload

from chellow.dloads import put_mem_val
from chellow.e.computer import (
    SiteSource,
    SupplySource,
    contract_func,
    displaced_era,
    forecast_date,
)
from chellow.models import Era, Session, Site, SiteEra, Source, Supply
from chellow.utils import c_months_u, ct_datetime_now, hh_max, hh_min


def totals_runner(mem_id, site_id, start_date):
    CATS = ("imp_net", "exp_net", "used", "displaced")
    QUANTS = (
        "kwh",
        "gbp",
        "p_per_kwh",
        "triad_gbp",
        "duos_gbp",
        "red_duos_gbp",
        "md_kw",
    )
    sess = None
    titles = ["Quantity"]
    titles.extend(CATS)
    log_list = []
    problem = ""

    jval = {"titles": titles, "rows": [], "status": "running", "log": log_list}
    try:
        sess = Session()
        site = Site.get_by_id(sess, site_id)

        def log(msg):
            return
            log_list.append(msg)
            put_mem_val(mem_id, jval)

        log("started thread")
        f_date = forecast_date()
        report_context = {}
        vals = dict((cat, defaultdict(int)) for cat in CATS)

        if start_date is None:
            now = ct_datetime_now()
            start_year = now.year - 1
            start_month = now.month
        else:
            start_year = start_date.year
            start_month = start_date.month

        for progress, (month_start, month_finish) in enumerate(
            c_months_u(start_year=start_year, start_month=start_month, months=12)
        ):
            jval["progress"] = progress
            put_mem_val(mem_id, jval)

            for era in sess.execute(
                select(Era)
                .join(SiteEra)
                .join(Supply)
                .join(Source)
                .where(
                    SiteEra.site == site,
                    Era.start_date <= month_finish,
                    or_(Era.finish_date == null(), Era.finish_date >= month_start),
                    Era.imp_mpan_core != null(),
                    Source.code.in_(("net", "gen-net")),
                )
                .options(
                    joinedload(Era.imp_supplier_contract),
                    joinedload(Era.exp_supplier_contract),
                    joinedload(Era.mop_contract),
                    joinedload(Era.dc_contract),
                )
            ).scalars():
                chunk_start = hh_max(era.start_date, month_start)
                chunk_finish = hh_min(era.finish_date, month_finish)

                log("about to call first supplysource")
                supply_ds = SupplySource(
                    sess,
                    chunk_start,
                    chunk_finish,
                    f_date,
                    era,
                    True,
                    report_context,
                )

                supplier_contract = era.imp_supplier_contract
                import_vb_function = contract_func(
                    report_context, supplier_contract, "virtual_bill"
                )
                if import_vb_function is None:
                    raise Exception(
                        "Can't find the import_virtual_bill function in the "
                        "supplier contract. "
                    )
                else:
                    import_vb_function(supply_ds)
                    v_bill = supply_ds.supplier_bill

                    if "problem" in v_bill and len(v_bill["problem"]) > 0:
                        raise Exception("Supplier Problem: " + v_bill["problem"])

                    try:
                        vals["imp_net"]["gbp"] += v_bill["net-gbp"]
                        if "triad-gbp" in v_bill:
                            vals["imp_net"]["triad_gbp"] += v_bill["triad-gbp"]
                        if "duos-red-gbp" in v_bill:
                            vals["imp_net"]["red_duos_gbp"] += v_bill["duos-red-gbp"]
                        vals["imp_net"]["duos_gbp"] += sum(
                            v
                            for k, v in v_bill.items()
                            if k.endswith("-gbp") and k.startswith("duos-")
                        )
                    except KeyError:
                        raise Exception(
                            f"For the era {era.id} the virtual bill {v_bill} "
                            f"from the contract {supplier_contract} does not "
                            f"contain the net-gbp key."
                        )

                dc_contract = era.dc_contract
                supply_ds.contract_func(dc_contract, "virtual_bill")(supply_ds)
                dc_bill = supply_ds.dc_bill
                vals["imp_net"]["gbp"] += dc_bill["net-gbp"]
                if "problem" in dc_bill and len(dc_bill["problem"]) > 0:
                    problem += "DC Problem: " + dc_bill["problem"]

                mop_contract = era.mop_contract
                mop_bill_function = supply_ds.contract_func(
                    mop_contract, "virtual_bill"
                )
                mop_bill_function(supply_ds)
                mop_bill = supply_ds.mop_bill
                vals["imp_net"]["gbp"] += mop_bill["net-gbp"]
                if "problem" in mop_bill and len(mop_bill["problem"]) > 0:
                    problem += "MOP Problem: " + mop_bill["problem"]

                exp_supplier_contract = era.exp_supplier_contract
                if exp_supplier_contract is not None:
                    log("about to call second supply source")
                    exp_supply_ds = SupplySource(
                        sess,
                        chunk_start,
                        chunk_finish,
                        f_date,
                        era,
                        False,
                        report_context,
                    )
                    exp_vb_function = contract_func(
                        report_context, exp_supplier_contract, "virtual_bill"
                    )
                    if exp_vb_function is None:
                        problem += (
                            "Can't find the export_virtual_bill function in the "
                            "supplier contract. "
                        )
                    else:
                        exp_vb_function(exp_supply_ds)
                        v_bill = exp_supply_ds.supplier_bill

                        if "problem" in v_bill and len(v_bill["problem"]) > 0:
                            problem += f"Export supplier Problem: {v_bill['problem']}"

                        try:
                            vals["exp_net"]["gbp"] += v_bill["net-gbp"]
                            if "triad-gbp" in v_bill:
                                vals["exp_net"]["triad_gbp"] += v_bill["triad-gbp"]
                            if "duos-red-gbp" in v_bill:
                                vals["exp_net"]["red_duos_gbp"] += v_bill[
                                    "duos-red-gbp"
                                ]
                            vals["exp_net"]["duos_gbp"] += sum(
                                v
                                for k, v in v_bill.items()
                                if k.endswith("-gbp") and k.startswith("duos-")
                            )
                        except KeyError:
                            problem += (
                                f"For the era {era.id} the virtual bill "
                                f"{v_bill} from the contract "
                                f"{supplier_contract} does not contain the "
                                f"net-gbp key."
                            )
                            problem += v_bill["problem"]

            disp_era = displaced_era(
                sess, report_context, site, month_start, month_finish, f_date
            )
            log("about to call sitesource")
            site_ds = SiteSource(
                sess,
                site,
                month_start,
                month_finish,
                f_date,
                report_context,
                disp_era,
                exclude_virtual=False,
            )

            if disp_era is not None:
                disp_supplier_contract = disp_era.imp_supplier_contract
                disp_vb_function = contract_func(
                    report_context, disp_supplier_contract, "displaced_virtual_bill"
                )
                if disp_vb_function is None:
                    raise Exception(
                        f"The supplier contract {disp_supplier_contract} "
                        "doesn't have the displaced_virtual_bill() function."
                    )
                disp_vb_function(site_ds)
                disp_supplier_bill = site_ds.supplier_bill

                try:
                    vals["displaced"]["gbp"] += disp_supplier_bill["net-gbp"]
                except KeyError:
                    disp_supplier_bill["problem"] += (
                        f"For the supply {site_ds.mpan_core} the virtual bill "
                        f"{disp_supplier_bill} from the contract "
                        f"{disp_supplier_contract} does not contain the net-gbp "
                        f"key."
                    )
                log(str(disp_supplier_bill))
                if "triad-gbp" in disp_supplier_bill:
                    vals["displaced"]["triad_gbp"] += disp_supplier_bill["triad-gbp"]
                if "duos-red-gbp" in disp_supplier_bill:
                    vals["displaced"]["red_duos_gbp"] += disp_supplier_bill[
                        "duos-red-gbp"
                    ]
                vals["displaced"]["duos_gbp"] += sum(
                    v
                    for k, v in disp_supplier_bill.items()
                    if k.endswith("-gbp") and k.startswith("duos-")
                )

            for hh in site.hh_data(
                sess, month_start, month_finish, exclude_virtual=True
            ):
                for cat in CATS:
                    vals_cat = vals[cat]
                    vals_cat["md_kw"] = max(vals_cat["md_kw"], hh[cat] * 2)
                    vals_cat["kwh"] += float(hh[cat])

        for quant in ("gbp", "triad_gbp", "duos_gbp", "red_duos_gbp"):
            vals["used"][quant] += vals["imp_net"][quant] + vals["displaced"][quant]

        for cat in CATS:
            vals_cat = vals[cat]
            if vals_cat["kwh"] > 0:
                vals_cat["p_per_kwh"] = vals_cat["gbp"] / vals_cat["kwh"] * 100
            else:
                vals_cat["p_per_kwh"] = 0

        FORMAT_LOOKUP = {
            "kwh": "{:,.0f}",
            "gbp": "{:,.2f}",
            "p_per_kwh": "{:,.2f}",
            "triad_gbp": "{:,.2f}",
            "duos_gbp": "{:,.2f}",
            "red_duos_gbp": "{:,.2f}",
            "md_kw": "{:,.2f}",
        }

        for quant in QUANTS:
            row = [quant]
            fmt = FORMAT_LOOKUP[quant]
            for cat in CATS:
                row.append(fmt.format(vals[cat][quant]))
            jval["rows"].append(row)

    except BaseException:
        msg = traceback.format_exc()
        jval["problem"] = msg
    finally:
        try:
            if sess is not None:
                sess.close()
        except BaseException:
            jval["problem"] = "\nProblem closing session."
        finally:
            jval["status"] = "finished"
            put_mem_val(mem_id, jval)



from flask import g

from flask_restx import Api, Resource, fields, inputs, reqparse

from sqlalchemy import null, select

from chellow.models import Channel, Era, HhDatum, Site


api = Api(
    version="1.0",
    title="Chellow API",
    description="Access Chellow data",
    endpoint="/api/v1",
    doc="/api/v1",
)

ns = api.namespace("v1", path="/api/v1/")

sites__model = ns.model(
    "Site",
    {
        "id": fields.Integer(example="884"),
        "code": fields.String(example="GGJ23"),
        "name": fields.String(example="Trowbridge Waterworks"),
    },
)

site__era__model = ns.model(
    "Era",
    {
        "id": fields.Integer(example="81"),
        "supply_id": fields.Integer(example="988"),
        "start_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "finish_date": fields.DateTime(example="2019-06-10T11:17:00+00:00"),
        "imp_mpan_core": fields.String,
        "exp_mpan_core": fields.String,
    },
)

site__site_era__model = ns.model(
    "SiteEra",
    {
        "id": fields.Integer(example="74"),
        "era": fields.Nested(site__era__model),
        "is_physical": fields.Boolean(example="true"),
    },
)

site__model = ns.model(
    "Site",
    {
        "id": fields.Integer(example="884"),
        "code": fields.String(example="GGJ23"),
        "name": fields.String(example="Trowbridge Waterworks"),
        "site_eras": fields.List(fields.Nested(site__site_era__model)),
    },
)

hh_data_model = ns.model(
    "HhData",
    {
        "from": fields.Integer,
        "to": fields.String,
        "name": fields.String,
    },
)

eras__site__model = ns.model(
    "Site",
    {
        "id": fields.Integer(example="884"),
        "code": fields.String(example="GGJ23"),
        "name": fields.String(example="Trowbridge Waterworks"),
    },
)

eras__site_eras__model = ns.model(
    "SiteEra",
    {
        "id": fields.Integer(example="74"),
        "site": fields.Nested(eras__site__model),
        "is_physical": fields.Boolean(example="true"),
    },
)

eras__mop_contract__model = ns.model(
    "MOP Contract",
    {
        "id": fields.Integer(example="821"),
        "name": fields.String(example="Marmaduke's Metering"),
    },
)

eras__dc_contract__model = ns.model(
    "DC Contract",
    {
        "id": fields.Integer(example="821"),
        "name": fields.String(example="Daisy's Data Collection"),
    },
)

pc__model = ns.model(
    "Profile Class",
    {
        "id": fields.Integer(example="8"),
        "code": fields.String(example="00"),
        "description": fields.String(example="Half-hourly"),
    },
)

era__participant__model = ns.model(
    "Participant",
    {
        "id": fields.Integer(example="91"),
        "code": fields.String(example="CUST"),
        "name": fields.String(example="Ingram Enterprises"),
    },
)

era__dno__model = ns.model(
    "DNO",
    {
        "id": fields.Integer(example="8"),
        "dno_code": fields.String(example="33"),
    },
)


era__mtc__model = ns.model(
    "Meter Timeswitch Class",
    {
        "id": fields.Integer(example="89"),
        "code": fields.String(example="800"),
        "description": fields.String(example="Unmetered"),
    },
)

cop__model = ns.model(
    "Code Of Practice",
    {
        "id": fields.Integer(example="89"),
        "code": fields.String(example="6c"),
        "description": fields.String(example="NHH"),
    },
)

eras__model = ns.model(
    "Eras",
    {
        "id": fields.Integer(example="712"),
        "supply_id": fields.Integer(example="349"),
        "site_eras": fields.List(fields.Nested(eras__site_eras__model)),
        "start_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "finish_date": fields.DateTime(example="2019-06-10T11:17:00+00:00"),
    },
)


era__clock_interval__model = ns.model(
    "Clock Interval",
    {
        "id": fields.Integer(example="920"),
        "day_of_week": fields.Integer(example="2"),
        "start_month": fields.Integer(example="10"),
        "start_day": fields.Integer(example="23"),
        "start_hour": fields.Integer(example="23"),
        "start_minute": fields.Integer(example="30"),
        "end_month": fields.Integer(example="12"),
        "end_day": fields.Integer(example="13"),
        "end_hour": fields.Integer(example="6"),
        "end_minute": fields.Integer(example="0"),
    },
)

era__tpr__model = ns.model(
    "Tariff Pattern Regime",
    {
        "id": fields.Integer(example="71"),
        "code": fields.String(example="00001"),
        "is_teleswitch": fields.Boolean(example="false"),
        "is_gmt": fields.Boolean(example="true"),
        "clock_intervals": fields.List(fields.Nested(era__clock_interval__model)),
    },
)


measurement_requirement__model = ns.model(
    "Measurement Requirement",
    {
        "id": fields.Integer(example="72"),
        "tpr": fields.Nested(era__tpr__model),
    },
)

ssc__model = ns.model(
    "Standard Settlement Configuration",
    {
        "id": fields.Integer(example="11"),
        "code": fields.String(example="0363"),
        "description": fields.String(example="Unrestricted"),
        "is_import": fields.Boolean(example="true"),
        "valid_from": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "valid_to": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "measurement_requirements": fields.List(
            fields.Nested(measurement_requirement__model)
        ),
    },
)

energisation_status__model = ns.model(
    "Energisation Status",
    {
        "id": fields.Integer(example="25"),
        "code": fields.String(example="E"),
        "description": fields.String(example="Energised"),
    },
)

era__llfc__model = ns.model(
    "Line Loss Factor Class",
    {
        "id": fields.Integer(example="54"),
        "code": fields.String(example="010"),
        "description": fields.String(example="HV Sub"),
    },
)

era__supplier_contract__model = ns.model(
    "Supplier Contract",
    {
        "id": fields.Integer(example="871"),
        "name": fields.String(example="High Charges"),
    },
)


era__mop_contract__model = ns.model(
    "MOP Contract",
    {
        "id": fields.Integer(example="821"),
        "name": fields.String(example="Marmaduke's Metering"),
    },
)


era__channel__model = ns.model(
    "Channel",
    {
        "id": fields.Integer(example="871"),
        "imp_related": fields.Boolean(example="true"),
        "channel_type": fields.String(example="ACTIVE"),
    },
)


era__model = ns.model(
    "Era",
    {
        "id": fields.Integer(example="712"),
        "supply_id": fields.Integer(example="349"),
        "site_eras": fields.List(fields.Nested(eras__site_eras__model)),
        "start_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "finish_date": fields.DateTime(example="2019-06-10T11:17:00+00:00"),
        "mop_contract": fields.Nested(era__mop_contract__model),
        "mop_account": fields.String(example="xc771"),
        "dc_contract": fields.Nested(eras__dc_contract__model),
        "dc_account": fields.String(example="703n"),
        "msn": fields.String(example="mn88hg2"),
        "pc": fields.Nested(pc__model),
        "mtc": fields.Nested(era__mtc__model),
        "cop": fields.Nested(cop__model),
        "ssc": fields.Nested(ssc__model),
        "energisation_status": fields.Nested(energisation_status__model),
        "properties": fields.String,
        "imp_mpan_core": fields.String(example="22 5478 7392 943"),
        "imp_llfc": fields.Nested(era__llfc__model),
        "imp_supplier_contract": fields.Nested(era__supplier_contract__model),
        "imp_supplier_account": fields.String(example="x66g"),
        "imp_sc": fields.Integer(example="700"),
        "exp_mpan_core": fields.String(example="22 8672 2951 081"),
        "exp_llfc": fields.Nested(era__llfc__model),
        "exp_supplier_contract": fields.Nested(era__supplier_contract__model),
        "exp_supplier_account": fields.String(example="h8892"),
        "exp_sc": fields.Integer(example="900"),
        "channels": fields.List(fields.Nested(era__channel__model)),
    },
)


channel__hh_datum__model = ns.model(
    "HH Datum",
    {
        "id": fields.Integer(example="871"),
        "channel_type": fields.String(example="ACTIVE"),
        "start_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
        "value": fields.String(example="0.01"),
        "last_modified_date": fields.DateTime(example="2019-05-18T15:17:00+00:00"),
    },
)


channel__model = ns.model(
    "Channel",
    {
        "id": fields.Integer(example="871"),
        "imp_related": fields.Boolean(example="true"),
        "channel_type": fields.String(example="ACTIVE"),
        "hh_data": fields.List(fields.Nested(channel__hh_datum__model)),
    },
)


@ns.route("/sites")
class SitesResource(Resource):
    @api.marshal_list_with(sites__model)
    def get(self):
        return g.sess.query(Site).order_by(Site.code).all()


@ns.route("/sites/<int:site_id>")
@ns.param("site_id", "The site identifier")
class SiteResource(Resource):
    @api.marshal_with(site__model)
    def get(self, site_id):
        return Site.get_by_id(g.sess, site_id)


channel_get_parser = reqparse.RequestParser()
channel_get_parser.add_argument(
    "from",
    type=inputs.datetime_from_iso8601,
    help="Beginning of date range for half-hourly data.",
)
channel_get_parser.add_argument(
    "to",
    type=inputs.datetime_from_iso8601,
    help="End of date range for half-hourly data.",
)


@ns.route("/eras/<int:era_id>")
class EraResource(Resource):
    @api.marshal_with(era__model)
    def get(self, era_id):
        return Era.get_by_id(g.sess, era_id)


channel_get_parser = reqparse.RequestParser()
channel_get_parser.add_argument(
    "from",
    type=inputs.datetime_from_iso8601,
    help="Beginning of date range for half-hourly data.",
)
channel_get_parser.add_argument(
    "to",
    type=inputs.datetime_from_iso8601,
    help="End of date range for half-hourly data.",
)


@ns.route("/channel/<int:channel_id>")
@ns.param("from", "YYYY-mm-dd HH:MM")
@ns.param("to", "YYYY-mm-dd HH:MM")
class ChannelResource(Resource):
    @api.marshal_with(channel__model)
    def get(self, channel_id):
        args = channel_get_parser.parse_args()
        channel = Channel.get_by_id(g.sess, channel_id)
        channel_j = {
            "id": channel.id,
            "era_id": channel.era_id,
            "imp_related": channel.imp_related,
            "channel_type": channel.channel_type,
        }
        channel_j["hh_data"] = g.sess.execute(
            select(HhDatum).where(
                HhDatum.channel == channel,
                HhDatum.start_date >= args["from"],
                HhDatum.start_date <= args["to"],
            )
        ).all()
        return channel_j


@ns.route("/eras")
class ErasResource(Resource):
    @api.marshal_list_with(eras__model)
    def get(self):
        return (
            g.sess.query(Era).filter(Era.finish_date == null()).order_by(Era.id).all()
        )



import csv
import os
import threading
import traceback
import zipfile
from io import StringIO

from flask import g, request

from sqlalchemy import or_
from sqlalchemy.sql.expression import null

import chellow.dloads
from chellow.models import Channel, Era, HhDatum, Session, Supply
from chellow.utils import (
    ct_datetime,
    hh_range,
    parse_mpan_core,
    req_bool,
    req_int,
    req_str,
    to_ct,
    to_utc,
)
from chellow.views import chellow_redirect


def content(
    start_date_ct,
    finish_date_ct,
    imp_related,
    channel_type,
    is_zipped,
    supply_id,
    mpan_cores,
    user,
):
    start_date, finish_date = to_utc(start_date_ct), to_utc(finish_date_ct)
    zf = sess = tf = None
    base_name = ["supplies_hh_data", finish_date_ct.strftime("%Y%m%d%H%M")]
    cache = {}
    try:
        sess = Session()
        supplies = (
            sess.query(Supply)
            .join(Era)
            .filter(
                or_(Era.finish_date == null(), Era.finish_date >= start_date),
                Era.start_date <= finish_date,
            )
            .order_by(Supply.id)
            .distinct()
        )
        if supply_id is not None:
            supply = Supply.get_by_id(sess, supply_id)
            supplies = supplies.filter(Supply.id == supply.id)
            first_era = (
                sess.query(Era)
                .filter(
                    Era.supply == supply,
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                    Era.start_date <= finish_date,
                )
                .order_by(Era.start_date)
                .first()
            )
            if first_era.imp_mpan_core is None:
                name_core = first_era.exp_mpan_core
            else:
                name_core = first_era.imp_mpan_core
            base_name.append("supply_" + name_core.replace(" ", "_"))

        if mpan_cores is not None:
            supplies = supplies.filter(
                or_(
                    Era.imp_mpan_core.in_(mpan_cores), Era.exp_mpan_core.in_(mpan_cores)
                )
            )
            base_name.append("filter")

        cf = StringIO()
        writer = csv.writer(cf, lineterminator="\n")
        titles = [
            "Import MPAN Core",
            "Export MPAN Core",
            "Import Related?",
            "Channel Type",
            "HH Start Clock-Time",
        ] + list(range(1, 51))
        writer.writerow(titles)
        titles_csv = cf.getvalue()
        cf.close()

        running_name, finished_name = chellow.dloads.make_names(
            "_".join(base_name) + (".zip" if is_zipped else ".csv"), user
        )
        if is_zipped:
            zf = zipfile.ZipFile(running_name, "w", zipfile.ZIP_DEFLATED)
        else:
            tf = open(running_name, mode="w", newline="")
            tf.write(titles_csv)

        for supply in supplies:
            cf = StringIO()
            writer = csv.writer(cf, lineterminator="\n")
            era = supply.find_era_at(sess, finish_date)
            if era is None:
                imp_mpan_core_str = exp_mpan_core_str = "NA"
            else:
                if era.imp_mpan_core is None:
                    imp_mpan_core_str = "NA"
                else:
                    imp_mpan_core_str = era.imp_mpan_core
                if era.exp_mpan_core is None:
                    exp_mpan_core_str = "NA"
                else:
                    exp_mpan_core_str = era.exp_mpan_core

            imp_related_str = "TRUE" if imp_related else "FALSE"

            hh_data = iter(
                sess.query(HhDatum)
                .join(Channel)
                .join(Era)
                .filter(
                    Era.supply == supply,
                    HhDatum.start_date >= start_date,
                    HhDatum.start_date <= finish_date,
                    Channel.imp_related == imp_related,
                    Channel.channel_type == channel_type,
                )
                .order_by(HhDatum.start_date)
            )
            datum = next(hh_data, None)

            row = []
            for current_date in hh_range(cache, start_date, finish_date):
                dt_ct = to_ct(current_date)
                if dt_ct.hour == 0 and dt_ct.minute == 0:
                    if len(row) > 0:
                        writer.writerow(row)
                    row = [
                        imp_mpan_core_str,
                        exp_mpan_core_str,
                        imp_related_str,
                        channel_type,
                        dt_ct.strftime("%Y-%m-%d"),
                    ]

                if datum is not None and datum.start_date == current_date:
                    row.append(datum.value)
                    datum = next(hh_data, None)
                else:
                    row.append(None)

            if len(row) > 0:
                writer.writerow(row)

            if is_zipped:
                fname = "_".join(
                    (imp_mpan_core_str, exp_mpan_core_str, str(supply.id) + ".csv")
                )
                zf.writestr(fname.encode("ascii"), titles_csv + cf.getvalue())
            else:
                tf.write(cf.getvalue())
            cf.close()

            # Avoid long-running transaction
            sess.rollback()

        if is_zipped:
            zf.close()
        else:
            tf.close()

    except BaseException:
        msg = traceback.format_exc()
        if is_zipped:
            zf.writestr("error.txt", msg)
            zf.close()
        else:
            tf.write(msg)
    finally:
        if sess is not None:
            sess.close()
        os.rename(running_name, finished_name)


def do_get(sess):
    return handle_request()


def do_post(sess):
    if "mpan_cores" in request.values:
        mpan_cores_str = req_str("mpan_cores")
        mpan_cores = mpan_cores_str.splitlines()
        if len(mpan_cores) == 0:
            mpan_cores = None
        else:
            for i in range(len(mpan_cores)):
                mpan_cores[i] = parse_mpan_core(mpan_cores[i])
    return handle_request(mpan_cores)


def handle_request(mpan_cores=None):
    start_year = req_int("start_year")
    start_month = req_int("start_month")
    start_day = req_int("start_day")
    start_date_ct = ct_datetime(start_year, start_month, start_day)

    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")
    finish_day = req_int("finish_day")
    finish_date_ct = ct_datetime(finish_year, finish_month, finish_day, 23, 30)

    imp_related = req_bool("imp_related")
    channel_type = req_str("channel_type")
    is_zipped = req_bool("is_zipped")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None
    user = g.user
    args = (
        start_date_ct,
        finish_date_ct,
        imp_related,
        channel_type,
        is_zipped,
        supply_id,
        mpan_cores,
        user,
    )
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from decimal import Decimal

from chellow.gas.engine import GDataSource, _find_hhs
from chellow.models import (
    Contract,
    GContract,
    GDn,
    GReadingFrequency,
    GUnit,
    MarketRole,
    Participant,
    Site,
    insert_g_reading_frequencies,
    insert_g_units,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_find_hhs_pairs_before_after_chunk_finish(mocker):
    """If there's a meter change, a pair can start after the end of the chunk.
    here we test the case for a pair before and after the chunk finish.
    """
    mocker.patch("chellow.gas.engine.find_cv", return_value=(39, 39))
    sess = mocker.Mock()
    caches = {}
    hist_g_era = mocker.Mock()
    hist_g_era.correction_factor = 1
    hist_g_era.g_unit = mocker.Mock()
    hist_g_era.g_unit.code = "M3"
    hist_g_era.g_unit.factor = 1

    pairs = [
        {"start-date": utc_datetime(2010, 1, 1), "units": 1},
        {"start-date": utc_datetime(2010, 1, 1, 0, 30), "units": 1},
    ]
    chunk_start = utc_datetime(2010, 1, 1)
    chunk_finish = utc_datetime(2010, 1, 1)
    g_cv_id = 0
    g_ldz_code = "SW"
    hhs = _find_hhs(
        sess, caches, hist_g_era, pairs, chunk_start, chunk_finish, g_cv_id, g_ldz_code
    )
    assert hhs == {
        utc_datetime(2010, 1, 1): {
            "unit_code": "M3",
            "unit_factor": 1.0,
            "units_consumed": 1,
            "correction_factor": 1.0,
            "calorific_value": 39,
            "avg_cv": 39,
        }
    }

    assert pairs == [
        {
            "start-date": utc_datetime(2010, 1, 1),
            "units": 1,
            "finish-date": utc_datetime(2010, 1, 1),
        },
    ]


def test_find_hhs_pair_after_chunk_finish(mocker):
    """If there's a meter change, a pair can start after the end of the chunk.
    Here we test for a single pair after the chunk finish.
    """
    mocker.patch("chellow.gas.engine.find_cv", return_value=(39, 39))
    sess = mocker.Mock()
    caches = {}
    hist_g_era = mocker.Mock()
    hist_g_era.correction_factor = 1
    hist_g_era.g_unit = mocker.Mock()
    hist_g_era.g_unit.code = "M3"
    hist_g_era.g_unit.factor = 1

    pairs = [{"start-date": utc_datetime(2010, 1, 1, 0, 30), "units": 1}]
    chunk_start = utc_datetime(2010, 1, 1)
    chunk_finish = utc_datetime(2010, 1, 1)
    g_cv_id = 0
    g_ldz_code = "SW"
    hhs = _find_hhs(
        sess, caches, hist_g_era, pairs, chunk_start, chunk_finish, g_cv_id, g_ldz_code
    )
    assert hhs == {
        utc_datetime(2010, 1, 1): {
            "unit_code": "M3",
            "unit_factor": 1.0,
            "units_consumed": 1,
            "correction_factor": 1.0,
            "calorific_value": 39,
            "avg_cv": 39,
        }
    }

    assert pairs == [
        {
            "start-date": utc_datetime(2010, 1, 1),
            "units": 1,
            "finish-date": utc_datetime(2010, 1, 1),
        },
    ]


def test_GDataSource_init(sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "None core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_cv_rate_script = {
        "cvs": {
            "EA": {
                1: {"applicable_at": utc_datetime(2020, 10, 3), "cv": 39.2000},
            }
        }
    }
    GContract.insert_industry(sess, "cv", "", {}, vf, None, g_cv_rate_script)
    ug_rate_script = {
        "ug_gbp_per_kwh": {"EA1": Decimal("40.1")},
    }
    GContract.insert_industry(sess, "ug", "", {}, vf, None, ug_rate_script)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        vf,
        None,
        bank_holiday_rate_script,
    )
    charge_script = """
from chellow.gas.engine import g_rates
from chellow.utils import reduce_bill_hhs


def virtual_bill_titles():
    return [
        'units_consumed', 'correction_factor', 'unit_code', 'unit_factor',
        'calorific_value', 'kwh', 'gas_rate', 'gas_gbp', 'ccl_rate',
        'standing_rate', 'standing_gbp', 'net_gbp', 'vat_gbp', 'gross_gbp',
        'problem']


def virtual_bill(ds):
    for hh in ds.hh_data:
        start_date = hh['start_date']
        bill_hh = ds.bill_hhs[start_date]
        bill_hh['units_consumed'] = hh['units_consumed']
        bill_hh['correction_factor'] = {hh['correction_factor']}
        bill_hh['unit_code'] = {hh['unit_code']}
        bill_hh['unit_factor'] = {hh['unit_factor']}
        bill_hh['calorific_value'] = {hh['calorific_value']}
        kwh = hh['kwh']
        bill_hh['kwh'] = kwh
        gas_rate = float(
            g_rates(ds.sess, ds.caches, db_id, start_date)['gas_rate'])
        bill_hh['gas_rate'] = {gas_rate}
        bill_hh['gas_gbp'] = gas_rate * kwh

        if hh['utc_is_month_end']:
            standing_rate = float(
                g_rates(
                    ds.sess, ds.caches, db_id, start_date)['standing_rate'])
            bill_hh['standing_rate'] = {standing_rate}
            bill_hh['standing_gbp'] = standing_rate
        if hh['utc_decimal_hour'] == 0:
            pass

        bill_hh['net_gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat_gbp'] = 0
        bill_hh['gross_gbp'] = bill_hh['net_gbp'] + bill_hh['vat_gbp']

    ds.bill = reduce_bill_hhs(ds.bill_hhs)
"""
    g_contract_rate_script = {
        "gas_rate": 0.1,
        "standing_rate": 0.1,
    }
    g_contract = GContract.insert(
        sess,
        False,
        "Fusion 2020",
        charge_script,
        {},
        vf,
        None,
        g_contract_rate_script,
    )
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    msn = "hgeu8rhg"
    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2010, 1, 1),
        None,
        msn,
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )
    sess.commit()

    g_era = g_supply.g_eras[0]
    start_date = to_utc(ct_datetime(2020, 1, 1))
    finish_date = to_utc(ct_datetime(2020, 1, 31, 23, 30))
    fdate = to_utc(ct_datetime(2020, 3, 1))
    caches = {}
    g_bill = None
    GDataSource(sess, start_date, finish_date, fdate, g_era, caches, g_bill)



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.models import Session
from chellow.utils import parse_mpan_core, to_utc


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)
        self.sheet = self.book.worksheets[1]

        self.last_line = None
        self._line_number = None
        self._title_line = None

    def get_cell(self, col, row):
        try:
            coordinates = f"{col}{row}"
            return self.sheet[coordinates]
        except IndexError:
            raise BadRequest(
                f"Can't find the cell {coordinates} on sheet {self.sheet}."
            )

    def get_int(self, col, row):
        return int(self.get_cell(col, row).value)

    def get_dec(self, col, row):
        cell = self.get_cell(col, row)
        try:
            return Decimal(str(cell.value))
        except InvalidOperation as e:
            raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")

    def get_str(self, col, row):
        return self.get_cell(col, row).value.strip()

    def get_ct_date(self, col, row):
        cell = self.get_cell(col, row)
        val = cell.value
        if not isinstance(val, Datetime):
            raise BadRequest(
                f"Problem reading {val} as a timestamp at {cell.coordinate}."
            )
        return val

    def get_start_date(self, col, row):
        dt = self.get_ct_date(col, row)
        return None if dt is None else to_utc(dt)

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        sess = None
        try:
            sess = Session()
            bills = []
            issue_date = self.get_start_date("C", 6)

            for row in range(12, len(self.sheet["A"]) + 1):
                val = self.get_cell("B", row).value
                if val is None or val == "":
                    break

                self._set_last_line(row, val)
                mpan_core = parse_mpan_core(str(self.get_int("B", row)))

                start_date = finish_date = self.get_start_date("F", row)
                activity_name_raw = self.get_str("G", row)
                activity_name = activity_name_raw.lower().replace(" ", "_")

                net_dec = self.get_dec("I", row)
                net = round(net_dec, 2)

                vat_dec = self.get_dec("J", row)
                vat = round(vat_dec, 2)

                gross_dec = self.get_dec("K", row)
                gross = round(gross_dec, 2)

                breakdown = {
                    "raw-lines": [],
                    "activity-name": [activity_name],
                    "activity-gbp": net,
                }

                bills.append(
                    {
                        "bill_type_code": "N",
                        "kwh": Decimal(0),
                        "vat": vat,
                        "net": net,
                        "gross": gross,
                        "reads": [],
                        "breakdown": breakdown,
                        "account": mpan_core,
                        "issue_date": issue_date,
                        "start_date": start_date,
                        "finish_date": finish_date,
                        "mpan_core": mpan_core,
                        "reference": "_".join(
                            (
                                start_date.strftime("%Y%m%d"),
                                finish_date.strftime("%Y%m%d"),
                                issue_date.strftime("%Y%m%d"),
                                mpan_core,
                            )
                        ),
                    }
                )
        except BadRequest as e:
            raise BadRequest(f"Row number: {row} {e.description}")
        finally:
            if sess is not None:
                sess.close()

        return bills



import csv
import os
import sys
import threading
import traceback
from collections import OrderedDict, defaultdict
from datetime import datetime as Datetime
from decimal import Decimal
from itertools import combinations

from flask import g, request

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

from werkzeug.exceptions import BadRequest

import chellow.dloads
import chellow.gas.engine
from chellow.models import GBatch, GBill, GEra, Session, Site, SiteGEra, User
from chellow.utils import csv_make_val, hh_max, hh_min, req_int, to_utc
from chellow.views import chellow_redirect


def content(g_batch_id, g_bill_id, user_id):
    forecast_date = to_utc(Datetime.max)
    report_context = {}
    sess = tmp_file = None
    try:
        sess = Session()
        user = User.get_by_id(sess, user_id)

        running_name, finished_name = chellow.dloads.make_names(
            "g_bill_check.csv", user
        )
        tmp_file = open(running_name, "w")
        csv_writer = csv.writer(tmp_file)
        if g_batch_id is not None:
            g_batch = GBatch.get_by_id(sess, g_batch_id)
            g_bills = (
                sess.query(GBill)
                .filter(GBill.g_batch == g_batch)
                .order_by(GBill.reference)
            )
        elif g_bill_id is not None:
            g_bill = GBill.get_by_id(sess, g_bill_id)
            g_bills = sess.query(GBill).filter(GBill.id == g_bill.id)
            g_batch = g_bill.g_batch

        g_contract = g_batch.g_contract

        vbf = chellow.gas.engine.g_contract_func(
            report_context, g_contract, "virtual_bill"
        )
        if vbf is None:
            raise BadRequest(
                f"The contract {g_contract.name} doesn't have a function "
                f"virtual_bill."
            )

        header_titles = [
            "batch",
            "bill_reference",
            "bill_type",
            "bill_start_date",
            "bill_finish_date",
            "mprn",
            "supply_name",
            "site_code",
            "site_name",
            "covered_start",
            "covered_finish",
            "covered_bill_ids",
        ]
        bill_titles = chellow.gas.engine.g_contract_func(
            report_context, g_contract, "virtual_bill_titles"
        )()

        titles = header_titles[:]
        for title in bill_titles:
            for prefix in ("covered_", "virtual_"):
                titles.append(prefix + title)
            if title.endswith("_gbp"):
                titles.append("difference_" + title)

        csv_writer.writerow(titles)

        g_bill_map = defaultdict(set, {})
        for b in g_bills:
            g_bill_map[b.g_supply.id].add(b.id)

        for g_supply_id, g_bill_ids in g_bill_map.items():
            while len(g_bill_ids) > 0:
                _process_g_bill_ids(
                    sess,
                    report_context,
                    g_bill_ids,
                    forecast_date,
                    bill_titles,
                    vbf,
                    titles,
                    csv_writer,
                )
    except BadRequest as e:
        tmp_file.write("Problem: " + e.description)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        tmp_file.write("Problem " + msg)
    finally:
        try:
            if sess is not None:
                sess.close()
        except BaseException:
            tmp_file.write("\nProblem closing session.")
        finally:
            tmp_file.close()
            os.rename(running_name, finished_name)


def _process_g_bill_ids(
    sess,
    report_context,
    g_bill_ids,
    forecast_date,
    bill_titles,
    vbf,
    titles,
    csv_writer,
):
    g_bill_id = list(sorted(g_bill_ids))[0]
    g_bill_ids.remove(g_bill_id)
    g_bill = sess.query(GBill).filter(GBill.id == g_bill_id).one()
    problem = ""
    g_supply = g_bill.g_supply
    read_dict = defaultdict(set)
    for g_read in g_bill.g_reads:
        if not all(
            g_read.msn == era.msn
            for era in g_supply.find_g_eras(sess, g_read.prev_date, g_read.pres_date)
        ):
            problem += (
                f"The MSN {g_read.msn} of the register read {g_read.id} doesn't match "
                f"the MSN of all the relevant eras."
            )

        for dt, typ in [
            (g_read.pres_date, g_read.pres_type),
            (g_read.prev_date, g_read.prev_type),
        ]:
            typ_set = read_dict[str(dt) + "-" + g_read.msn]
            typ_set.add(typ)
            if len(typ_set) > 1:
                problem += f" Reads taken on {dt} have differing read types."

    vals = {
        "covered_vat_gbp": Decimal("0.00"),
        "covered_net_gbp": Decimal("0.00"),
        "covered_gross_gbp": Decimal("0.00"),
        "covered_kwh": Decimal(0),
        "covered_start": g_bill.start_date,
        "covered_finish": g_bill.finish_date,
        "covered_bill_ids": [],
    }

    covered_primary_bill = None
    enlarged = True

    while enlarged:
        enlarged = False
        covered_bills = OrderedDict(
            (b.id, b)
            for b in sess.query(GBill)
            .filter(
                GBill.g_supply == g_supply,
                GBill.start_date <= vals["covered_finish"],
                GBill.finish_date >= vals["covered_start"],
            )
            .order_by(GBill.issue_date.desc(), GBill.start_date)
        )

        num_covered = None
        while num_covered != len(covered_bills):
            num_covered = len(covered_bills)
            for a, b in combinations(tuple(covered_bills.values()), 2):
                if all(
                    (
                        a.start_date == b.start_date,
                        a.finish_date == b.finish_date,
                        a.kwh == -1 * b.kwh,
                        a.net == -1 * b.net,
                        a.vat == -1 * b.vat,
                        a.gross == -1 * b.gross,
                    )
                ):
                    for gb_id in a.id, b.id:
                        del covered_bills[gb_id]
                        if gb_id in g_bill_ids:
                            g_bill_ids.remove(gb_id)
                    break

        for covered_bill in covered_bills.values():
            if covered_primary_bill is None and len(covered_bill.g_reads) > 0:
                covered_primary_bill = covered_bill
            if covered_bill.start_date < vals["covered_start"]:
                vals["covered_start"] = covered_bill.start_date
                enlarged = True
                break
            if covered_bill.finish_date > vals["covered_finish"]:
                vals["covered_finish"] = covered_bill.finish_date
                enlarged = True
                break

    if len(covered_bills) == 0:
        return

    for covered_bill in covered_bills.values():
        if covered_bill.id in g_bill_ids:
            g_bill_ids.remove(covered_bill.id)
        vals["covered_bill_ids"].append(covered_bill.id)
        bdown = covered_bill.make_breakdown()
        vals["covered_kwh"] += covered_bill.kwh
        vals["covered_net_gbp"] += covered_bill.net
        vals["covered_vat_gbp"] += covered_bill.vat
        vals["covered_gross_gbp"] += covered_bill.gross
        for title in bill_titles:
            k = "covered_" + title
            v = bdown.get(title)

            if v is not None:
                if isinstance(v, list):
                    if k not in vals:
                        vals[k] = set()
                    vals[k].update(set(v))
                else:
                    try:
                        vals[k] += v
                    except KeyError:
                        vals[k] = v
                    except TypeError:
                        raise BadRequest(
                            f"Problem with bill {g_bill.id} and key {k} and value {v} "
                            f"for existing {vals[k]}"
                        )

            if title in (
                "correction_factor",
                "calorific_value",
                "unit_code",
                "unit_factor",
            ):
                if k not in vals:
                    vals[k] = set()
                for g_read in covered_bill.g_reads:
                    if title in ("unit_code", "unit_factor"):
                        g_unit = g_read.g_unit
                        if title == "unit_code":
                            v = g_unit.code
                        else:
                            v = g_unit.factor
                    else:
                        v = getattr(g_read, title)
                    vals[k].add(v)

    for g_era in (
        sess.query(GEra)
        .filter(
            GEra.g_supply == g_supply,
            GEra.start_date <= vals["covered_finish"],
            or_(GEra.finish_date == null(), GEra.finish_date >= vals["covered_start"]),
        )
        .distinct()
    ):
        site = (
            sess.query(Site)
            .join(SiteGEra)
            .filter(SiteGEra.is_physical == true(), SiteGEra.g_era == g_era)
            .one()
        )

        chunk_start = hh_max(vals["covered_start"], g_era.start_date)
        chunk_finish = hh_min(vals["covered_finish"], g_era.finish_date)

        data_source = chellow.gas.engine.GDataSource(
            sess,
            chunk_start,
            chunk_finish,
            forecast_date,
            g_era,
            report_context,
            covered_primary_bill,
        )

        vbf(data_source)

        for k, v in data_source.bill.items():
            vk = "virtual_" + k
            try:
                if isinstance(v, set):
                    vals[vk].update(v)
                else:
                    vals[vk] += v
            except KeyError:
                vals[vk] = v
            except TypeError as detail:
                raise BadRequest(f"For key {vk} and value {v}. {detail}")

    if g_bill.id not in covered_bills.keys():
        g_bill = covered_bills[sorted(covered_bills.keys())[0]]

    vals["batch"] = g_bill.g_batch.reference
    vals["bill_reference"] = g_bill.reference
    vals["bill_type"] = g_bill.bill_type.code
    vals["bill_start_date"] = g_bill.start_date
    vals["bill_finish_date"] = g_bill.finish_date
    vals["mprn"] = g_supply.mprn
    vals["supply_name"] = g_supply.name
    vals["site_code"] = site.code
    vals["site_name"] = site.name

    for k, v in vals.items():
        if k == "covered_bill_ids":
            vals[k] = " | ".join(str(b) for b in v)
        else:
            vals[k] = csv_make_val(v)

    for i, title in enumerate(titles):
        if title.startswith("difference_"):
            try:
                covered_val = float(vals[titles[i - 2]])
                virtual_val = float(vals[titles[i - 1]])
                vals[title] = covered_val - virtual_val
            except KeyError:
                vals[title] = None

    csv_writer.writerow(
        [(vals.get(k) if vals.get(k) is not None else "") for k in titles]
    )


def do_get(sess):
    if "g_batch_id" in request.values:
        g_batch_id = req_int("g_batch_id")
        g_bill_id = None
    elif "g_bill_id" in request.values:
        g_bill_id = req_int("g_bill_id")
        g_batch_id = None
    else:
        raise BadRequest("The bill check needs a g_batch_id or g_bill_id.")

    threading.Thread(target=content, args=(g_batch_id, g_bill_id, g.user.id)).start()
    return chellow_redirect("/downloads", 303)



import atexit
import collections
import csv
import threading
import traceback
from datetime import datetime as Datetime
from decimal import Decimal
from io import BytesIO

from dateutil.relativedelta import relativedelta

from pypdf import PdfReader

import requests

from sqlalchemy import or_, select
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

import xlrd

from zish import dumps, loads

from chellow.models import (
    Contract,
    RateScript,
    Session,
    get_non_core_contract_id,
)
from chellow.rate_server import download
from chellow.utils import HH, ct_datetime, hh_format, to_ct, to_utc, utc_datetime_now

MAXIMA = [
    {
        "maximum": 15,
        "start": to_utc(ct_datetime(2020, 6, 25)),
        "finish": to_utc(ct_datetime(2020, 8, 31, 23, 30)),
        "reference": "CMP345",
    },
    {
        "maximum": 20,
        "start": to_utc(ct_datetime(2022, 1, 17)),
        "finish": to_utc(ct_datetime(2022, 3, 31, 23, 30)),
        "reference": "CMP381",
    },
    {
        "maximum": 40,
        "start": to_utc(ct_datetime(2022, 10, 6)),
        "finish": to_utc(ct_datetime(2023, 3, 31, 23, 30)),
        "reference": "CMP395",
    },
]


def hh(data_source, run="RF"):
    try:
        bsuos_cache = data_source.caches["bsuos"][run]
    except KeyError:
        try:
            top_cache = data_source.caches["bsuos"]
        except KeyError:
            top_cache = data_source.caches["bsuos"] = {}

        try:
            bsuos_cache = top_cache[run]
        except KeyError:
            bsuos_cache = top_cache[run] = {}

    for h in data_source.hh_data:
        try:
            h["bsuos-rate"] = bsuos_rate = bsuos_cache[h["start-date"]]
        except KeyError:
            h_start = h["start-date"]
            db_id = get_non_core_contract_id("bsuos")

            rates = data_source.hh_rate(db_id, h_start)

            if h_start >= to_utc(ct_datetime(2023, 4, 1)):
                bsuos_price = float(rates["rate_gbp_per_mwh"])
            else:
                bsuos_rates = rates["rates_gbp_per_mwh"]

                maxi = 0
                for max_dict in MAXIMA:
                    if max_dict["start"] <= h_start <= max_dict["finish"]:
                        maxi = max_dict["maximum"]
                        break

                try:
                    bsuos_price_dict = bsuos_rates[key_format(h_start)]
                    bsuos_price = _find_price(run, bsuos_price_dict, maxi)
                except KeyError:
                    ds = rates.values()
                    bsuos_price = sum(_find_price(run, d, maxi) for d in ds) / len(ds)

            h["bsuos-rate"] = bsuos_rate = bsuos_cache[h_start] = bsuos_price / 1000

        h["bsuos-kwh"] = h["nbp-kwh"]
        h["bsuos-gbp"] = h["nbp-kwh"] * bsuos_rate


def _find_price(run, prices, maxi):
    try:
        price = prices[run]
    except KeyError:
        try:
            price = prices["RF"]
        except KeyError:
            try:
                price = prices["SF"]
            except KeyError:
                price = prices["II"]

    return min(float(price), maxi)


def key_format(dt):
    return dt.strftime("%d %H:%M")


bsuos_importer = None


def _save_cache(sess, cache):
    for yr, yr_cache in cache.items():
        for month, (rs, rates, rts) in tuple(yr_cache.items()):
            rs.script = dumps(rates)
            sess.commit()
            del yr_cache[month]


class BsuosImporter(threading.Thread):
    def __init__(self):
        super(BsuosImporter, self).__init__(name="BSUoS Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=500)
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.PROXY_HOST_KEY = "proxy.host"
        self.PROXY_PORT_KEY = "proxy.port"

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                sess = self.global_alert = s = None
                try:
                    sess = Session()
                    self.log("Starting to check BSUoS rates.")
                    contract = Contract.get_non_core_by_name(sess, "bsuos")
                    props = contract.make_properties()
                    s = requests.Session()
                    s.verify = False
                    if props.get("enabled", False):
                        urls = set(props.get("urls", []))
                        if props.get("discover_urls", False):
                            sess.rollback()  # Avoid long-running transaction
                            urls.update(_discover_urls(self.log, s))

                        url_list = sorted(urls)
                        self.log(f"List of URLs to process: {url_list}")
                        for url in url_list:
                            _process_url(self.log, sess, url, contract, s)
                    else:
                        self.log(
                            "The automatic importer is disabled. To enable it, edit "
                            "the contract properties to set 'enabled' to True."
                        )
                except BaseException:
                    self.log(f"Outer problem {traceback.format_exc()}")
                    self.global_alert = (
                        f"There's a problem with the "
                        f"<a href='/non_core_contracts/{contract.id}'>BSUoS automatic "
                        f"importer</a>."
                    )
                    sess.rollback()
                finally:
                    if sess is not None:
                        sess.close()
                    self.lock.release()
                    self.log("Finished checking BSUoS rates.")

            self.going.wait(60 * 60 * 24)
            self.going.clear()


def _find_file_type(disp):
    # Content-Disposition: form-data; name="fieldName"; filename="filename.jpg"
    filetype = "csv"
    if disp is not None:
        fields = dict(v.strip().lower().split("=") for v in disp.split(";") if "=" in v)
        if "filename" in fields:
            filetype = fields["filename"].strip('"').split(".")[-1]

    return filetype


def _process_url(logger, sess, url, contract, s):
    logger(f"Checking to see if there's any new data at {url}")
    sess.rollback()  # Avoid long-running transaction
    res = s.get(url)
    content_disposition = res.headers.get("Content-Disposition")
    logger(f"Received {res.status_code} {res.reason} {content_disposition}")
    cache = {}
    parsed_rows = []

    filetype = _find_file_type(content_disposition)
    if filetype == "csv":
        reader = csv.reader(res.text.splitlines())
        next(reader)  # Skip titles

        for row in reader:
            date_str = row[0]
            date = Datetime.strptime(date_str, "%d/%m/%Y")
            period_str = row[1]
            period = int(period_str)
            price_str = row[2]
            price = Decimal(price_str)
            run = row[5]
            parsed_rows.append((date, period, price, run))

    elif filetype == "xls":
        book = xlrd.open_workbook(file_contents=res.content)
        sheet = book.sheet_by_index(0)

        for row_index in range(1, sheet.nrows):
            row = sheet.row(row_index)

            date_val = row[0].value
            if isinstance(date_val, float):
                date = Datetime(*xlrd.xldate_as_tuple(date_val, book.datemode))
            elif isinstance(date_val, str):
                separator = date_val[2]
                fmat = separator.join(("%d", "%m", "%Y"))
                date = Datetime.strptime(date_val, fmat)
            else:
                raise BadRequest(f"Type of date field {date_val} not recognized.")

            period = int(row[1].value)
            price = Decimal(str(row[2].value))
            run = row[5].value
            parsed_rows.append((date, period, price, run))
    else:
        raise BadRequest(f"The file extension {filetype} is not recognised.")

    for date, period, price, run in parsed_rows:
        hh_date_ct = to_ct(date)
        hh_date_ct += relativedelta(minutes=30 * (period - 1))
        hh_date = to_utc(hh_date_ct)
        if hh_date >= to_utc(ct_datetime(2023, 4, 1)):
            continue

        try:
            rs, rates, rts = cache[hh_date.year][hh_date.month]
        except KeyError:
            _save_cache(sess, cache)

            try:
                yr_cache = cache[hh_date.year]
            except KeyError:
                yr_cache = cache[hh_date.year] = {}

            rs = (
                sess.query(RateScript)
                .filter(
                    RateScript.contract == contract,
                    RateScript.start_date <= hh_date,
                    or_(
                        RateScript.finish_date == null(),
                        RateScript.finish_date >= hh_date,
                    ),
                )
                .first()
            )
            while rs is None:
                logger(f"There's no rate script at {hh_format(hh_date)}.")
                latest_rs = (
                    sess.query(RateScript)
                    .filter(RateScript.contract == contract)
                    .order_by(RateScript.start_date.desc())
                    .first()
                )
                contract.update_rate_script(
                    sess,
                    latest_rs,
                    latest_rs.start_date,
                    latest_rs.start_date + relativedelta(months=2) - HH,
                    loads(latest_rs.script),
                )
                new_rs_start = latest_rs.start_date + relativedelta(months=1)
                contract.insert_rate_script(sess, new_rs_start, {})
                sess.commit()
                logger(f"Added a rate script starting at {hh_format(new_rs_start)}.")

                rs = (
                    sess.query(RateScript)
                    .filter(
                        RateScript.contract == contract,
                        RateScript.start_date <= hh_date,
                        or_(
                            RateScript.finish_date == null(),
                            RateScript.finish_date >= hh_date,
                        ),
                    )
                    .first()
                )

            rates = loads(rs.script)
            try:
                rts = rates["rates_gbp_per_mwh"]
            except KeyError:
                rts = rates["rates_gbp_per_mwh"] = {}
            yr_cache[hh_date.month] = rs, rates, rts

        key = key_format(hh_date)
        try:
            existing = rts[key]
        except KeyError:
            existing = rts[key] = {}

        if run not in existing:
            existing[run] = price
            logger(f"Added rate at {hh_format(hh_date)} for run {run}.")

    _save_cache(sess, cache)


def _discover_urls(logger, s):
    host = "https://www.nationalgrideso.com"
    page = (
        f"{host}/industry-information/charging/"
        "balancing-services-use-system-bsuos-charges"
    )
    logger(f"Searching for URLs on {page}")
    urls = set()
    res = s.get(page)
    src = res.text

    # <a href="/document/102541/download" target="_blank">Current RF BSUoS Data </a>

    for pref in ("RF", "SF", "II"):
        pidx = src.find(f'" target="_blank">Current {pref} BSUoS Data <')
        aidx = src.rfind('<a href="', 0, pidx)
        urls.add(host + src[aidx + 9 : pidx])
    return urls


def get_importer():
    return bsuos_importer


def startup():
    global bsuos_importer
    bsuos_importer = BsuosImporter()
    bsuos_importer.start()


@atexit.register
def shutdown():
    if bsuos_importer is not None:
        bsuos_importer.stop()


def find_rate(file_name, file_like, rate_index):
    rate_script = {"a_file_name": file_name}
    rates = []

    reader = PdfReader(file_like)
    for page in reader.pages:
        for line in page.extract_text().splitlines():
            if line.startswith("BSUoS Tariff £/MWh"):
                for token in line.split():
                    if token[0] == "£" and token[1] != "/":
                        rates.append(Decimal(token[1:]))

    rate_script["rate_gbp_per_mwh"] = rates[rate_index]
    return rate_script


def rate_server_import(sess, log, set_progress, s, paths):
    log("Starting to check for new BSUoS spreadsheets")

    year_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year_str, utility, rate_type, file_name = path
            if utility == "electricity" and rate_type == "bsuos":
                year = int(year_str)
                try:
                    fl_entries = year_entries[year]
                except KeyError:
                    fl_entries = year_entries[year] = {}

                fl_entries[file_name] = url

    for year, year_pdfs in sorted(year_entries.items()):
        year_start = to_utc(ct_datetime(year, 4, 1))
        oct_start = to_utc(ct_datetime(year, 10, 1))
        contract = Contract.get_non_core_by_name(sess, "bsuos")
        if year_start < contract.start_rate_script.start_date:
            continue
        rs_1 = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == year_start,
            )
        ).scalar_one_or_none()
        if rs_1 is None:
            rs_1 = contract.insert_rate_script(sess, year_start, {})

        rs_2 = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == oct_start,
            )
        ).scalar_one_or_none()
        if rs_2 is None:
            rs_2 = contract.insert_rate_script(sess, oct_start, {})

        if len(year_pdfs) > 0:
            file_name, url = sorted(year_pdfs.items())[-1]

            rs_1_script = rs_1.make_script()
            if rs_1_script.get("a_file_name") != file_name:
                rs_1.update(find_rate(file_name, BytesIO(download(s, url)), 0))
                log(f"Updated BSUoS rate script for {hh_format(year_start)}")

            rs_2_script = rs_2.make_script()
            if rs_2_script.get("a_file_name") != file_name:
                rs_2.update(find_rate(file_name, BytesIO(download(s, url)), 1))
                log(f"Updated BSUoS rate script for {hh_format(oct_start)}")

    log("Finished BSUoS spreadsheets")
    sess.commit()



import pytest

from werkzeug.exceptions import BadRequest

from chellow.e.bill_parsers.haven_csv import _process_READING, _process_line, _to_date
from chellow.utils import ct_datetime, to_utc


def test_process_line_unknown_code(mocker):
    code = "BOB"
    row = []
    headers = {}

    with pytest.raises(BadRequest, match=code):
        _process_line(code, row, headers)


def test_process_line_SUMMARY():
    code = "SUMMARY"
    row = []
    headers = {}

    _process_line(code, row, headers)


def test_to_date():
    date_str = "20200430"
    row = [date_str]
    dt = _to_date(row, 0)
    assert dt == to_utc(ct_datetime(2020, 4, 30))


def test_process_READING_unknown_unit():
    units = "BOB"
    row = [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        "20200401",  # 11
        0,
        "E",
        "20200430",
        0,
        "E",
        0,
        units,
    ]
    headers = {}

    with pytest.raises(BadRequest, match=units):
        _process_READING(row, headers)


def test_process_READING_success():
    row = [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        "20200401",  # 11
        0,
        "E",
        "20200430",
        0,
        "E",
        0,
        "KWH",
    ]
    headers = {"reads": []}

    _process_READING(row, headers)



from collections import defaultdict
from decimal import Decimal


from werkzeug.exceptions import BadRequest

from chellow.edi_lib import parse_edi, to_date, to_decimal, to_finish_date
from chellow.utils import HH


read_type_map = {
    "00": "N",
    "09": "N3",
    "04": "C",
    "02": "E",
    "11": "E3",
    "01": "EM",
    "03": "W",
    "06": "X",
    "05": "CP",
    "12": "IF",
}


TCOD_MAP = {
    "140114": ("reconciliation-gbp", None, None),
    "255204": ("meter-rental-gbp", "meter-rental-rate", "meter-rental-days"),
    "345065": ("summer-weekend-gbp", "summer-weekend-rate", "summer-weekend-kwh"),
    "350293": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "425779": ("ro-gbp", "ro-rate", "ro-kwh"),
    "534342": ("reconciliation-gbp", None, None),
    "583174": ("meter-rental-gbp", "meter-rental-rate", "meter-rental-days"),
    "584867": ("aahedc-gbp", "aahedc-rate", "aahedc-kwh"),
    "946827": ("meter-rental-gbp", "meter-rental-rate", "meter-rental-days"),
    "989534": ("bsuos-gbp", "bsuos-rate", "bsuos-kwh"),
    "117220": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "579387": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "558147": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "030025": ("ccl-gbp", "ccl-rate", "ccl-kwh"),
    "066540": ("ccl-gbp", "ccl-rate", "ccl-kwh"),
    "154164": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "281170": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "342094": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "378809": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "574015": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "810016": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "839829": ("cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-kwh"),
    "068476": ("day-gbp", "day-rate", "day-kwh"),
    "133186": ("day-gbp", "day-rate", "day-kwh"),
    "400434": ("day-gbp", "day-rate", "day-kwh"),
    "219182": (
        "duos-availability-gbp",
        "duos-availability-rate",
        "duos-availability-kva",
    ),
    "144424": (
        "duos-excess-availability-gbp",
        "duos-excess-availability-rate",
        "duos-excess-availability-kva",
    ),
    "986159": ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days"),
    "838286": ("duos-reactive-gbp", "duos-reactive-rate", "duos-reactive-kvarh"),
    "242643": ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days"),
    "257304": ("duos-amber-gbp", "duos-amber-rate", "duos-amber-kwh"),
    "661440": ("duos-amber-gbp", "duos-amber-rate", "duos-red-kwh"),
    "257305": ("duos-green-gbp", "duos-green-rate", "duos-green-kwh"),
    "257303": ("duos-red-gbp", "duos-red-rate", "duos-red-kwh"),
    "661439": ("duos-red-gbp", "duos-red-rate", "duos-red-kwh"),
    "661441": ("duos-red-gbp", "duos-red-rate", "duos-red-kwh"),
    "563023": ("ebrs-gbp", None, "ebrs-kwh"),
    "823408": ("ebrs-gbp", None, "ebrs-kwh"),
    "871593": ("ebrs-gbp", "ebrs-rate", "ebrs-kwh"),
    "873894": ("ebrs-gbp", "ebrs-rate", "ebrs-kwh"),
    "309707": ("fit-gbp", "fit-rate", "fit-kwh"),
    "994483": ("reconciliation-gbp", None, None),
    "544936": ("meter-rental-gbp", "meter-rental-rate", "meter-rental-days"),
    "265091": ("night-gbp", "night-rate", "night-kwh"),
    "483457": ("peak-gbp", "peak-rate", "peak-kwh"),
    "975901": ("peak-shoulder-gbp", "peak-shoulder-rate", "peak-shoulder-kwh"),
    "364252": ("ro-gbp", "ro-rate", "ro-kwh"),
    "378246": ("ro-gbp", "ro-rate", "ro-kwh"),
    "632209": ("summer-night-gbp", "summer-night-rate", "summer-night-kwh"),
    "663682": ("summer-weekday-gbp", "summer-weekday-rate", "summer-weekday-kwh"),
    "299992": ("summer-weekend-gbp", "summer-weekend-rate", "summer-weekend-kwh"),
    "447769": ("triad-gbp", "triad-rate", "triad-kw"),
    "647721": ("triad-gbp", "triad-rate", "triad-kw"),
    "276631": ("triad-gbp", "triad-rate", "triad-kw"),
    "220894": ("winter-night-gbp", "winter-night-rate", "winter-night-kwh"),
    "264929": ("winter-weekday-gbp", "winter-weekday-rate", "winter-weekday-kwh"),
    "638187": ("winter-weekend-gbp", "winter-weekend-rate", "winter-weekend-kwh"),
    "700285": ("duo-fixed-gbp", "duos-fixed-rate", "duos-fixed-days"),
}


def _process_BCD(elements, headers):
    issue_date = to_date(elements["IVDT"][0])
    reference = elements["INVN"][0]
    bill_type_code = elements["BTCD"][0]

    headers["issue_date"] = issue_date
    headers["bill_type_code"] = bill_type_code
    headers["reference"] = reference


def _process_CCD1(elements, headers):
    pres_read_date = to_finish_date(elements["PRDT"][0])

    prev_read_date = to_finish_date(elements["PVDT"][0])

    m = elements["MLOC"][0]
    mpan = " ".join((m[13:15], m[15:18], m[18:], m[:2], m[2:6], m[6:10], m[10:13]))

    prrd = elements["PRRD"]
    if len(prrd) < 4:
        return
    pres_read_type = read_type_map[prrd[1]]
    prev_read_type = read_type_map[prrd[3]]

    coefficient = Decimal(elements["ADJF"][1]) / Decimal(100000)
    pres_reading_value = Decimal(prrd[0])
    prev_reading_value = Decimal(prrd[2])
    msn = elements["MTNR"][0]
    tpr_code = elements["TMOD"][0]
    if tpr_code == "kW":
        units = "kW"
        tpr_code = None
    elif tpr_code == "kVA":
        units = "kVA"
        tpr_code = None
    else:
        units = "kWh"

    try:
        reads = headers["reads"]
    except KeyError:
        reads = headers["reads"] = []

    reads.append(
        {
            "msn": msn,
            "mpan": mpan,
            "coefficient": coefficient,
            "units": units,
            "tpr_code": tpr_code,
            "prev_date": prev_read_date,
            "prev_value": prev_reading_value,
            "prev_type_code": prev_read_type,
            "pres_date": pres_read_date,
            "pres_value": pres_reading_value,
            "pres_type_code": pres_read_type,
        }
    )


def _process_CCD2(elements, headers):
    breakdown = defaultdict(int)

    element_code = elements["TCOD"][0]
    try:
        eln_gbp, eln_rate, eln_cons = TCOD_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TCOD_MAP.")

    m = elements["MLOC"][0]
    mpan_core = " ".join((m[:2], m[2:6], m[6:10], m[10:]))

    cons = elements["CONS"]
    kwh = Decimal("0")
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons) / Decimal("1000")
        if eln_gbp == "duos-availability-gbp":
            breakdown[eln_cons] = [el_cons]
        else:
            breakdown[eln_cons] = kwh = el_cons

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"]) / Decimal("100000")
        breakdown[eln_rate] = [rate]

    start_date = to_date(elements["CSDT"][0])
    headers["bill_start_date"] = start_date

    finish_date = to_date(elements["CEDT"][0]) - HH
    headers["bill_finish_date"] = finish_date

    if "CTOT" in elements:
        net = Decimal("0.00") + to_decimal(elements["CTOT"]) / Decimal("100")
    else:
        net = Decimal("0.00")

    breakdown[eln_gbp] = net

    headers["mpan_core"] = mpan_core

    try:
        reads = headers["reads"]
        del headers["reads"][:]
    except KeyError:
        reads = []

    return {
        "bill_type_code": headers["bill_type_code"],
        "reference": headers["reference"] + "_" + eln_gbp[:-4],
        "issue_date": headers["issue_date"],
        "mpan_core": mpan_core,
        "account": mpan_core,
        "start_date": start_date,
        "finish_date": finish_date,
        "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0"),
        "net": net,
        "vat": Decimal("0.00"),
        "gross": net,
        "breakdown": breakdown,
        "reads": reads,
    }


def _process_CCD3(elements, headers):
    breakdown = defaultdict(int)

    element_code = elements["TCOD"][0]
    try:
        eln_gbp, eln_rate, eln_cons = TCOD_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TCOD_MAP.")

    m = elements["MLOC"][0]
    mpan_core = " ".join((m[:2], m[2:6], m[6:10], m[10:]))

    cons = elements["CONS"]
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons) / Decimal("1000")
        breakdown[eln_cons] = kwh = el_cons
    else:
        kwh = Decimal("0")

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"]) / Decimal("100000")
        breakdown[eln_rate] = [rate]

    start_date = to_date(elements["CSDT"][0])
    headers["bill_start_date"] = start_date

    finish_date = to_date(elements["CEDT"][0]) - HH
    headers["bill_finish_date"] = finish_date

    if "CTOT" in elements:
        net = Decimal("0.00") + to_decimal(elements["CTOT"]) / Decimal("100")
    else:
        net = Decimal("0.00")

    breakdown[eln_gbp] = net

    headers["mpan_core"] = mpan_core

    try:
        reads = headers["reads"]
        del headers["reads"][:]
    except KeyError:
        reads = []

    return {
        "bill_type_code": headers["bill_type_code"],
        "issue_date": headers["issue_date"],
        "reference": headers["reference"],
        "mpan_core": mpan_core,
        "account": mpan_core,
        "start_date": start_date,
        "finish_date": finish_date,
        "net": net,
        "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0"),
        "vat": Decimal("0.00"),
        "gross": net,
        "breakdown": breakdown,
        "reads": reads,
    }


def _process_CCD4(elements, headers):
    breakdown = defaultdict(int)

    element_code = elements["TCOD"][0]
    try:
        eln_gbp, eln_rate, eln_cons = TCOD_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TCOD_MAP.")

    m = elements["MLOC"][0]
    mpan_core = " ".join((m[:2], m[2:6], m[6:10], m[10:]))

    cons = elements["CONS"]
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons, "1000")
        breakdown[eln_cons] = kwh = el_cons

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"], "100000")
        breakdown[eln_rate] = [rate]

    start_date = to_date(elements["CSDT"][0])
    headers["bill_start_date"] = start_date

    finish_date = to_date(elements["CEDT"][0]) - HH
    headers["bill_finish_date"] = finish_date

    if "CTOT" in elements:
        net = Decimal("0.00") + to_decimal(elements["CTOT"], "100")
    else:
        net = Decimal("0.00")

    breakdown[eln_gbp] = net

    headers["mpan_core"] = mpan_core

    try:
        reads = headers["reads"]
        del headers["reads"][:]
    except KeyError:
        reads = []

    return {
        "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0.00"),
        "reference": headers["reference"] + "_" + eln_gbp[:-4],
        "issue_date": headers["issue_date"],
        "mpan_core": mpan_core,
        "account": mpan_core,
        "start_date": start_date,
        "finish_date": finish_date,
        "net": net,
        "vat": Decimal("0.00"),
        "gross": net,
        "breakdown": breakdown,
        "reads": reads,
        "bill_type_code": headers["bill_type_code"],
    }


def _process_MHD(elements, headers):
    message_type = elements["TYPE"][0]
    if message_type == "UTLBIL":
        pass


def _process_MTR(elements, headers):
    headers.clear()


def _process_VAT(elements, headers):
    vat = Decimal("0.00") + to_decimal(elements["UVTT"]) / Decimal("100")

    return {
        "bill_type_code": headers["bill_type_code"],
        "account": headers["mpan_core"],
        "mpan_core": headers["mpan_core"],
        "reference": headers["reference"] + "_vat",
        "issue_date": headers["issue_date"],
        "start_date": headers["bill_start_date"],
        "finish_date": headers["bill_finish_date"],
        "kwh": Decimal("0.00"),
        "net": Decimal("0.00"),
        "vat": vat,
        "gross": vat,
        "breakdown": {},
        "reads": [],
    }


def _process_NOOP(elements, headers):
    pass


CODE_FUNCS = {
    "BCD": _process_BCD,
    "BTL": _process_NOOP,
    "CCD1": _process_CCD1,
    "CCD2": _process_CCD2,
    "CCD3": _process_CCD3,
    "CCD4": _process_CCD4,
    "CDT": _process_NOOP,
    "CLO": _process_NOOP,
    "DNA": _process_NOOP,
    "END": _process_NOOP,
    "FIL": _process_NOOP,
    "MAN": _process_NOOP,
    "MHD": _process_MHD,
    "MTR": _process_MTR,
    "SDT": _process_NOOP,
    "STX": _process_NOOP,
    "TYP": _process_NOOP,
    "TTL": _process_NOOP,
    "VAT": _process_VAT,
    "VTS": _process_NOOP,
}


class Parser:
    def __init__(self, f):
        self.edi_str = str(f.read(), "utf-8", errors="ignore")
        self.line_number = None

    def make_raw_bills(self):
        bills = []
        headers = {}
        bill = None
        for self.line_number, line, seg_name, elements in parse_edi(self.edi_str):
            try:
                func = CODE_FUNCS[seg_name]
            except KeyError:
                raise BadRequest(f"Code {seg_name} not recognized.")

            try:
                bill = func(elements, headers)
            except BadRequest as e:
                raise BadRequest(
                    f"{e.description} on line {self.line_number} line {line} "
                    f"seg_name {seg_name} elements {elements}"
                )
            except BaseException as e:
                raise BadRequest(
                    f"{e} on line {self.line_number} line {line} "
                    f"seg_name {seg_name} elements {elements}"
                ) from e

            if "breakdown" in headers:
                headers["breakdown"]["raw-lines"].append(line)

            if bill is not None:
                bills.append(bill)

        return bills



import csv
import os
import threading
import traceback
from datetime import datetime as Datetime

from dateutil.relativedelta import relativedelta

from flask import g

import pytz

from sqlalchemy.sql.expression import null, or_, true

from werkzeug.exceptions import BadRequest

import chellow.computer
from chellow.models import Era, Session, Site, SiteEra, Supply
from chellow.utils import HH, csv_make_val, hh_format, hh_max, hh_min, req_bool, req_int
from chellow.views import chellow_redirect


def content(
    start_year,
    start_month,
    start_day,
    finish_year,
    finish_month,
    finish_day,
    is_import,
    supply_id,
    user,
):
    caches = {}
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            "daily_supplier_virtual_bill.csv", user
        )
        f = open(running_name, mode="w", newline="")
        writer = csv.writer(f, lineterminator="\n")
        start_date = Datetime(start_year, start_month, start_day, tzinfo=pytz.utc)
        finish_date = (
            Datetime(finish_year, finish_month, finish_day, tzinfo=pytz.utc)
            + relativedelta(days=1)
            - HH
        )

        supply = Supply.get_by_id(sess, supply_id)
        forecast_date = chellow.computer.forecast_date()
        day_start = start_date
        header_titles = [
            "MPAN Core",
            "Site Code",
            "Site Name",
            "Account",
            "From",
            "To",
            "Is Forecast?",
        ]

        bill_titles = []
        # Find titles
        for era in sess.query(Era).filter(
            Era.supply == supply,
            Era.start_date <= finish_date,
            or_(Era.finish_date == null(), Era.finish_date >= start_date),
        ):
            if is_import:
                cont = era.imp_supplier_contract
            else:
                cont = era.exp_supplier_contract

            for title in chellow.computer.contract_func(
                caches, cont, "virtual_bill_titles"
            )():
                if title not in bill_titles:
                    bill_titles.append(title)

            ssc = era.ssc
            if ssc is not None:
                for mr in ssc.measurement_requirements:
                    for suffix in ("-kwh", "-rate", "-gbp"):
                        title = mr.tpr.code + suffix
                        if title not in bill_titles:
                            bill_titles.append(title)

        writer.writerow(header_titles + bill_titles)

        while not day_start > finish_date:
            day_finish = day_start + relativedelta(days=1) - HH

            for era in supply.find_eras(sess, day_start, day_finish):
                chunk_start = hh_max(era.start_date, day_start)
                chunk_finish = hh_min(era.finish_date, day_finish)

                ss = chellow.computer.SupplySource(
                    sess,
                    chunk_start,
                    chunk_finish,
                    forecast_date,
                    era,
                    is_import,
                    caches,
                )

                site = (
                    sess.query(Site)
                    .join(SiteEra)
                    .filter(SiteEra.era == era, SiteEra.is_physical == true())
                    .one()
                )
                row = [
                    ss.mpan_core,
                    site.code,
                    site.name,
                    ss.supplier_account,
                    hh_format(ss.start_date),
                    hh_format(ss.finish_date),
                    ss.years_back > 0,
                ]

                chellow.computer.contract_func(
                    caches, ss.supplier_contract, "virtual_bill"
                )(ss)
                bill = ss.supplier_bill
                for title in bill_titles:
                    if title in bill:
                        row.append(csv_make_val(bill[title]))
                        del bill[title]
                    else:
                        row.append("")

                for k in sorted(bill.keys()):
                    row.append(k)
                    row.append(csv_make_val(bill[k]))
                writer.writerow(row)

            day_start += relativedelta(days=1)
    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    start_year = req_int("start_year")
    start_month = req_int("start_month")
    start_day = req_int("start_day")

    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")
    finish_day = req_int("finish_day")

    is_import = req_bool("is_import")
    supply_id = req_int("supply_id")

    threading.Thread(
        target=content,
        args=(
            start_year,
            start_month,
            start_day,
            finish_year,
            finish_month,
            finish_day,
            is_import,
            supply_id,
            g.user,
        ),
    ).start()
    return chellow_redirect("/downloads", 303)



import csv
import os
import sys
import threading
import traceback

from flask import g

import chellow.dloads
from chellow.models import Session, Site
from chellow.utils import c_months_u, req_int, req_str, to_ct
from chellow.views import chellow_redirect


def content(start_date, finish_date, site_id, typ, user):
    sess = f = writer = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            "site_hh_data_" + to_ct(start_date).strftime("%Y%m%d%H%M") + ".csv", user
        )
        f = open(running_name, mode="w", newline="")
        writer = csv.writer(f, lineterminator="\n")
        writer.writerow(
            ("Site Code", "Type", "HH Start Clock-Time") + tuple(map(str, range(1, 51)))
        )
        site = Site.get_by_id(sess, site_id)
        line = None
        for hh in site.hh_data(sess, start_date, finish_date):
            hh_start_ct = to_ct(hh["start_date"])
            if (hh_start_ct.hour, hh_start_ct.minute) == (0, 0):
                if line is not None:
                    writer.writerow(line)
                line = [site.code, typ, hh_start_ct.strftime("%Y-%m-%d")]
            line.append(str(hh[typ]))
        if line is not None:
            writer.writerow(line)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    months = req_int("months")
    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")

    month_list = list(
        c_months_u(finish_year=finish_year, finish_month=finish_month, months=months)
    )
    start_date, finish_date = month_list[0][0], month_list[-1][-1]

    typ = req_str("type")
    site_id = req_int("site_id")
    args = (start_date, finish_date, site_id, typ, g.user)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from chellow import create_app, get_importer_modules
from chellow.models import Pc, Session


def test_create_app(mocker, fresh_db):
    mock_startups = []
    for importer_module in get_importer_modules():
        mock_startup = importer_module.startup = mocker.Mock()
        mock_startups.append(mock_startup)

    mock_module = mocker.patch("chellow.chellow.dloads")
    mock_startup = mock_module.startup = mocker.Mock()
    mock_startups.append(mock_startup)

    create_app()

    for mock_startup in mock_startups:
        mock_startup.assert_called_once()

    sess = Session()
    Pc.get_by_code(sess, "00")
    sess.close()



from chellow.e.bill_parsers.settlement_dc_stark_xlsx import Parser


def test(sess):
    f = open(
        "test/e/bill_parsers/test_bill_parser_settlement_dc_stark_xlsx/"
        "bills.settlement.dc.stark.xlsx",
        mode="rb",
    )
    parser = Parser(f)
    parser.make_raw_bills()



from io import StringIO

from sqlalchemy.orm.session import Session

from chellow.models import User, UserRole
from chellow.reports.report_41 import _make_eras, content
from chellow.utils import utc_datetime


def test_eras(mocker):
    """If there's a meter change, a pair can start after the end of the chunk.
    here we test the case for a pair before and after the chunk finish.
    """
    sess = Session()
    year_start = utc_datetime(2010, 4, 1)
    year_finish = utc_datetime(2011, 3, 31, 23, 30)
    supply_id = None
    eras = _make_eras(sess, year_start, year_finish, supply_id)
    desired = "".join(
        (
            "SELECT era.id AS era_id, era.supply_id AS era_supply_id, "
            "era.start_date AS era_start_date, "
            "era.finish_date AS era_finish_date, "
            "era.mop_contract_id AS era_mop_contract_id, "
            "era.mop_account AS era_mop_account, "
            "era.dc_contract_id AS era_dc_contract_id, "
            "era.dc_account AS era_dc_account, "
            "era.msn AS era_msn, "
            "era.pc_id AS era_pc_id, "
            "era.mtc_participant_id AS era_mtc_participant_id, "
            "era.cop_id AS era_cop_id, "
            "era.comm_id AS era_comm_id, "
            "era.ssc_id AS era_ssc_id, "
            "era.energisation_status_id AS era_energisation_status_id, "
            "era.properties AS era_properties, "
            "era.imp_mpan_core AS era_imp_mpan_core, "
            "era.imp_llfc_id AS era_imp_llfc_id, "
            "era.imp_supplier_contract_id AS era_imp_supplier_contract_id, "
            "era.imp_supplier_account AS era_imp_supplier_account, "
            "era.imp_sc AS era_imp_sc, "
            "era.exp_mpan_core AS era_exp_mpan_core, "
            "era.exp_llfc_id AS era_exp_llfc_id, "
            "era.exp_supplier_contract_id AS era_exp_supplier_contract_id, "
            "era.exp_supplier_account AS era_exp_supplier_account, "
            "era.exp_sc AS era_exp_sc \n"
            "FROM era JOIN supply ON supply.id = era.supply_id "
            "JOIN source ON source.id = supply.source_id "
            "JOIN pc ON pc.id = era.pc_id \n"
            "WHERE era.start_date <= :start_date_1 AND "
            "(era.finish_date IS NULL OR era.finish_date >= :finish_date_1) "
            "AND source.code IN (__[POSTCOMPILE_code_1]) AND pc.code = :code_2 "
            "ORDER BY supply.id"
        )
    )
    print(desired)
    assert str(eras) == desired


def test_content(mocker, sess):
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    sess.commit()

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_41.open", return_value=mock_file)
    mocker.patch("chellow.reports.report_41.make_names", return_value=("a", "b"))
    mocker.patch("chellow.reports.report_41.os.rename")

    year = 2010
    supply_id = None
    user_id = user.id
    content(year, supply_id, user_id)

    actual = mock_file.getvalue()
    expected = [
        "Site Code",
        "Site Name",
        "Supply Name",
        "Source",
        "Generator Type",
        "Import MPAN Core",
        "Import T1 Date",
        "Import T1 MSP kW",
        "Import T1 Status",
        "Import T1 LAF",
        "Import T1 GSP kW",
        "Import T2 Date",
        "Import T2 MSP kW",
        "Import T2 Status",
        "Import T2 LAF",
        "Import T2 GSP kW",
        "Import T3 Date",
        "Import T3 MSP kW",
        "Import T3 Status",
        "Import T3 LAF",
        "Import T3 GSP kW",
        "Import GSP kW",
        "Import Rate GBP / kW",
        "Import GBP",
        "Export MPAN Core",
        "Export T1 Date",
        "Export T1 MSP kW",
        "Export T1 Status",
        "Export T1 LAF",
        "Export T1 GSP kW",
        "Export T2 Date",
        "Export T2 MSP kW",
        "Export T2 Status",
        "Export T2 LAF",
        "Export T2 GSP kW",
        "Export T3 Date",
        "Export T3 MSP kW",
        "Export T3 Status",
        "Export T3 LAF",
        "Export T3 GSP kW",
        "Export GSP kW",
        "Export Rate GBP / kW",
        "Export GBP",
    ]
    print(actual)
    assert actual == ",".join(expected) + "\n"



import decimal
import itertools

from werkzeug.exceptions import BadRequest

from chellow.utils import HH, parse_mpan_core, utc_datetime


def create_parser(reader, mpan_map):
    return StarkDf2HhParser(reader, mpan_map)


class StarkDf2HhParser:
    def __init__(self, reader, mpan_map):
        self.line = self.line_number = None
        self.reader = zip(itertools.count(1), reader)

        self.line_number, self.line = next(self.reader)
        if self.line.strip().upper() != "#F2":
            raise BadRequest("The first line must be '#F2'.")
        self.sensor_map = {
            1: "ACTIVE",
            2: "ACTIVE",
            3: "REACTIVE_IMP",
            4: "REACTIVE_EXP",
        }

    def __iter__(self):
        return self

    def __next__(self):
        local_datum = None
        try:
            while local_datum is None:
                self.line_number, self.line = next(self.reader)
                lline = self.line.strip().upper()
                if lline.startswith("#O"):
                    self.core = parse_mpan_core(lline[2:])
                elif lline.startswith("#S"):
                    sensor = int(lline[2:].strip())
                    try:
                        self.channel_type = self.sensor_map[sensor]
                    except KeyError:
                        raise BadRequest(
                            "The sensor number must be between 1 and 4 " "inclusive."
                        )
                elif lline.startswith("#F2") or len(lline) == 0:
                    continue
                else:
                    fields = [f.strip() for f in lline.split(",")]
                    if len(fields) != 4:
                        raise BadRequest(
                            "There should be 4 comma separated values, but I "
                            "found " + str(len(fields)) + "."
                        )

                    d_day, d_month, d_year = map(int, fields[0].split("/"))
                    time_fields = tuple(map(int, fields[1].split(":")))
                    if len(time_fields) > 2 and time_fields[2] != 0:
                        raise BadRequest(
                            "The number of seconds (if present) must always " "be zero."
                        )
                    start_date = (
                        utc_datetime(
                            d_year, d_month, d_day, time_fields[0], time_fields[1]
                        )
                        - HH
                    )

                    try:
                        value = decimal.Decimal(fields[2])
                    except ValueError:
                        raise BadRequest("Problem parsing the value: " + fields[2])
                    status = fields[3][-1]
                    local_datum = {
                        "mpan_core": self.core,
                        "channel_type": self.channel_type,
                        "start_date": start_date,
                        "value": value,
                        "status": status,
                    }
            return local_datum
        except BadRequest as e:
            e.description = "".join(
                (
                    "Problem at line number: ",
                    str(self.line_number),
                    ": ",
                    self.line,
                    ": ",
                    e.description,
                )
            )
            raise e

    def close(self):
        self.reader.close()

    def get_status(self):
        return "Reached line number: " + str(self.line_number)



from decimal import Decimal
from io import BytesIO


from chellow.e.bill_parsers.engie_edi import (
    CODE_FUNCS,
    Parser,
    _process_BCD,
    _process_CCD1,
    _process_CCD2,
    _process_CCD3,
    _process_MHD,
    _process_NOOP,
    _process_VAT,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_CODE_FUNCS():
    assert CODE_FUNCS["BCD"] == _process_BCD
    assert CODE_FUNCS["BTL"] == _process_NOOP
    assert CODE_FUNCS["END"] == _process_NOOP
    assert CODE_FUNCS["TTL"] == _process_NOOP
    assert CODE_FUNCS["VTS"] == _process_NOOP


def test_process_CCD1(mocker):
    elements = {
        "CCDE": ["1", "f"],
        "TCOD": ["584867", "AAHEDC"],
        "TMOD": ["1", "2", "3", "4"],
        "MTNR": ["rikt8"],
        "MLOC": ["alskdj"],
        "PRDT": ["180901"],
        "PVDT": ["180801"],
        "NDRP": ["x"],
        "PRRD": ["9876", "00", 78649, "00"],
        "CONS": ["77", "kWh", "R"],
        "CONB": ["9850", "kWh", "R"],
        "ADJF": ["x", "688", "R"],
        "CONA": ["743", "kWh", "R"],
        "BPRI": ["895"],
        "NUCT": ["68349", "kWh", "R"],
        "CSDT": ["180901"],
        "CEDT": ["180801"],
        "CPPU": ["7003"],
        "CTOT": ["78679", "R"],
        "TSUP": ["jf"],
        "VATC": ["ig"],
        "VATP": ["77"],
        "MSAD": ["fk", "fdk"],
    }
    headers = {}
    _process_CCD1(elements, headers)
    expected_headers = {
        "reads": [
            {
                "msn": "rikt8",
                "mpan": "   al skdj  ",
                "coefficient": Decimal("0.00688"),
                "units": "kWh",
                "tpr_code": "1",
                "prev_date": utc_datetime(2018, 8, 1, 22, 30),
                "prev_value": Decimal("78649"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2018, 9, 1, 22, 30),
                "pres_value": Decimal("9876"),
                "pres_type_code": "N",
            }
        ]
    }
    assert headers == expected_headers


def test_process_CCD2_duos_availability(mocker):
    elements = {
        "CCDE": ["2", "ADD"],
        "TCOD": ["219182", "DUoS Availability"],
        "TMOD": [],
        "MTNR": [],
        "MLOC": ["22767395756734"],
        "PRDT": [],
        "PVDT": [],
        "NDRP": [],
        "PRRD": [],
        "CONS": ["877457492", "KWH"],
        "CONB": [],
        "ADJF": ["UG"],
        "CONA": [],
        "BPRI": ["974"],
        "NUCT": ["877457492", "KWH"],
        "CSDT": ["191001"],
        "CEDT": ["191101"],
        "CPPU": ["748"],
        "CTOT": ["76981"],
    }
    reference = "kdhgsf"
    issue_date = utc_datetime(2019, 4, 1)
    headers = {"reference": reference, "issue_date": issue_date, "bill_type_code": "N"}
    bill = _process_CCD2(elements, headers)
    expected_headers = {
        "bill_type_code": "N",
        "reference": reference,
        "issue_date": issue_date,
        "mpan_core": "22 7673 9575 6734",
        "bill_start_date": to_utc(ct_datetime(2019, 10, 1)),
        "bill_finish_date": to_utc(ct_datetime(2019, 10, 31, 23, 30)),
    }
    expected_bill = {
        "bill_type_code": "N",
        "reference": "kdhgsf_duos-availability",
        "issue_date": issue_date,
        "mpan_core": "22 7673 9575 6734",
        "account": "22 7673 9575 6734",
        "start_date": utc_datetime(2019, 9, 30, 23, 0),
        "finish_date": utc_datetime(2019, 10, 31, 23, 30),
        "kwh": Decimal("0"),
        "net": Decimal("769.81"),
        "vat": Decimal("0.00"),
        "gross": Decimal("769.81"),
        "breakdown": {
            "duos-availability-kva": [Decimal("877457.492")],
            "duos-availability-rate": [Decimal("0.00974")],
            "duos-availability-gbp": Decimal("769.81"),
        },
        "reads": [],
    }

    assert headers == expected_headers
    print(bill)
    print(expected_bill)
    assert bill == expected_bill
    assert isinstance(bill["kwh"], Decimal)
    assert isinstance(bill["net"], Decimal)
    assert isinstance(bill["vat"], Decimal)
    assert isinstance(bill["gross"], Decimal)
    assert str(bill["net"]) == str(expected_bill["net"])


def test_process_CCD3(mocker):
    elements = {
        "CCDE": ["3", "ADD"],
        "TCOD": ["584867", "AAHEDC"],
        "TMOD": [],
        "MTNR": [],
        "MLOC": ["22767395756734"],
        "PRDT": [],
        "PVDT": [],
        "NDRP": [],
        "PRRD": [],
        "CONS": ["877457492", "KWH"],
        "CONB": [],
        "ADJF": ["UG"],
        "CONA": [],
        "BPRI": ["974"],
        "NUCT": ["877457492", "KWH"],
        "CSDT": ["191001"],
        "CEDT": ["191101"],
        "CPPU": ["748"],
        "CTOT": ["76981"],
    }

    issue_date = utc_datetime(2019, 9, 3)
    reference = "hgtuer8"
    headers = {"issue_date": issue_date, "reference": reference, "bill_type_code": "N"}
    _process_CCD3(elements, headers)
    expected_headers = {
        "mpan_core": "22 7673 9575 6734",
        "bill_start_date": to_utc(ct_datetime(2019, 10, 1)),
        "bill_finish_date": to_utc(ct_datetime(2019, 10, 31, 23, 30)),
        "issue_date": issue_date,
        "reference": reference,
        "bill_type_code": "N",
    }
    assert headers == expected_headers


def test_process_CCD3_ro(mocker):
    elements = {
        "CCDE": ["3", "ADD"],
        "TCOD": ["425779", "RO Mutualisation"],
        "TMOD": [],
        "MTNR": [],
        "MLOC": ["22767395756734"],
        "PRDT": [],
        "PVDT": [],
        "NDRP": [],
        "PRRD": [],
        "CONS": ["", ""],
        "CONB": [],
        "ADJF": ["UG"],
        "CONA": [],
        "BPRI": ["974"],
        "NUCT": ["877457492", "KWH"],
        "CSDT": ["191001"],
        "CEDT": ["191101"],
        "CPPU": ["748"],
        "CTOT": ["76981"],
    }
    issue_date = utc_datetime(2019, 9, 3)
    reference = "hgtuer8"
    headers = {"issue_date": issue_date, "reference": reference, "bill_type_code": "N"}
    _process_CCD3(elements, headers)
    expected_headers = {
        "mpan_core": "22 7673 9575 6734",
        "bill_start_date": to_utc(ct_datetime(2019, 10, 1)),
        "bill_finish_date": to_utc(ct_datetime(2019, 10, 31, 23, 30)),
        "issue_date": issue_date,
        "reference": reference,
        "bill_type_code": "N",
    }
    assert headers == expected_headers


def test_process_CCD2_no_CTOT(mocker):
    elements = {
        "CCDE": ["2", "ADD"],
        "TCOD": ["584867", "AAHEDC"],
        "TMOD": [],
        "MTNR": [],
        "MLOC": ["22767395756734"],
        "PRDT": [],
        "PVDT": [],
        "NDRP": [],
        "PRRD": [],
        "CONS": ["877457492", "KWH"],
        "CONB": [],
        "ADJF": ["UG"],
        "CONA": [],
        "BPRI": ["974"],
        "NUCT": ["877457492", "KWH"],
        "CSDT": ["191001"],
        "CEDT": ["191101"],
        "CPPU": ["748"],
    }
    headers = {
        "reference": "shkfsd",
        "bill_type_code": "N",
        "issue_date": utc_datetime(2019, 9, 3),
    }
    bill = _process_CCD2(elements, headers)
    assert str(bill["net"]) == "0.00"


def test_process_CCD2_blank_CONS(mocker):
    elements = {
        "CCDE": ["2", "ADD"],
        "TCOD": ["584867", "AAHEDC"],
        "TMOD": [],
        "MTNR": [],
        "MLOC": ["22767395756734"],
        "PRDT": [],
        "PVDT": [],
        "NDRP": [],
        "PRRD": [],
        "CONS": [""],
        "CONB": [],
        "ADJF": ["UG"],
        "CONA": [],
        "BPRI": ["974"],
        "NUCT": ["877457492", "KWH"],
        "CSDT": ["191001"],
        "CEDT": ["191101"],
        "CPPU": ["748"],
    }
    headers = {
        "reference": "hgdertk",
        "bill_type_code": "N",
        "issue_date": utc_datetime(2019, 9, 3),
    }
    _process_CCD2(elements, headers)


def test_process_segment_CCD2_blank_ro(mocker):
    elements = {
        "CCDE": ["2", "ADD"],
        "TCOD": ["378246", "Ro"],
        "TMOD": [],
        "MTNR": [],
        "MLOC": ["22767395756734"],
        "PRDT": [],
        "PVDT": [],
        "NDRP": [],
        "PRRD": [],
        "CONS": [""],
        "CONB": [],
        "ADJF": ["UG"],
        "CONA": [],
        "BPRI": ["974"],
        "NUCT": [],
        "CSDT": ["191001"],
        "CEDT": ["191101"],
        "CPPU": ["748"],
        "CTOT": ["76981"],
    }
    reference = "kdhgsf"
    issue_date = utc_datetime(2019, 4, 1)
    headers = {"reference": reference, "issue_date": issue_date, "bill_type_code": "N"}
    bill = _process_CCD2(elements, headers)
    expected_headers = {
        "bill_type_code": "N",
        "reference": reference,
        "issue_date": issue_date,
        "mpan_core": "22 7673 9575 6734",
        "bill_start_date": to_utc(ct_datetime(2019, 10, 1)),
        "bill_finish_date": to_utc(ct_datetime(2019, 10, 31, 23, 30)),
    }
    expected_bill = {
        "bill_type_code": "N",
        "reference": "kdhgsf_ro",
        "issue_date": issue_date,
        "mpan_core": "22 7673 9575 6734",
        "account": "22 7673 9575 6734",
        "start_date": utc_datetime(2019, 9, 30, 23, 0),
        "finish_date": utc_datetime(2019, 10, 31, 23, 30),
        "kwh": Decimal("0.00"),
        "net": Decimal("769.81"),
        "vat": Decimal("0.00"),
        "gross": Decimal("769.81"),
        "breakdown": {
            "ro-rate": [Decimal("0.00974")],
            "ro-gbp": Decimal("769.81"),
        },
        "reads": [],
    }

    assert headers == expected_headers
    assert bill == expected_bill
    assert isinstance(bill["kwh"], Decimal)
    assert isinstance(bill["net"], Decimal)
    assert isinstance(bill["vat"], Decimal)
    assert isinstance(bill["gross"], Decimal)
    assert str(bill["net"]) == str(expected_bill["net"])


def test_process_MHD(mocker):
    elements = {
        "MSRF": ["ref"],
        "TYPE": ["a", "56"],
    }
    headers = {}
    _process_MHD(elements, headers)


def test_process_VAT(mocker):
    elements = {"UVTT": ["0"]}
    headers = {
        "mpan_core": "22 7673 9575 6734",
        "reference": "xx2",
        "issue_date": to_utc(ct_datetime(2020, 3, 1)),
        "bill_start_date": to_utc(ct_datetime(2020, 1, 1)),
        "bill_finish_date": to_utc(ct_datetime(2020, 1, 31)),
        "bill_type_code": "N",
    }
    _process_VAT(elements, headers)


def test_make_raw_bills(mocker):
    edi_file = BytesIO()
    parser = Parser(edi_file)
    assert parser.make_raw_bills() == []



import csv
from io import StringIO

from chellow.models import ReportRun, Site
from chellow.reports.report_asset_comparison import _process_sites


def test_process_sites(sess):
    site_code = "109"
    Site.insert(sess, site_code, "A Site")
    report_run = ReportRun.insert(sess, "asset", None, "asset", {})
    sess.commit()

    asset_str = "\n".join(("code,,,status",))
    file_like = StringIO(asset_str)
    output = StringIO()
    writer = csv.writer(output, lineterminator="\n")
    props = {"asset_comparison": {"ignore_site_codes": []}}

    _process_sites(sess, file_like, writer, props, report_run.id)

    assert output.getvalue() == "site_code,asset_status,chellow_status\n"



import csv
import os
import sys
import threading
import traceback


from flask import g

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

from werkzeug.exceptions import BadRequest

import chellow.dloads
from chellow.e.computer import contract_func, forecast_date
from chellow.gas.engine import GDataSource
from chellow.models import GContract, GEra, Session, Site, SiteGEra
from chellow.utils import (
    c_months_u,
    hh_format,
    hh_max,
    hh_min,
    make_val,
    req_date,
    req_int,
    to_ct,
)
from chellow.views import chellow_redirect


def content(start_date, finish_date, g_contract_id, user):
    report_context = {}
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            "gas_virtual_bills.csv", user
        )
        f = open(running_name, mode="w", newline="")
        writer = csv.writer(f, lineterminator="\n")

        g_contract = GContract.get_by_id(sess, g_contract_id)
        forecast_dt = forecast_date()

        start_date_ct, finish_date_ct = to_ct(start_date), to_ct(finish_date)

        month_pairs = c_months_u(
            start_year=start_date_ct.year,
            start_month=start_date_ct.month,
            finish_year=finish_date_ct.year,
            finish_month=finish_date_ct.month,
        )

        bill_titles = contract_func(report_context, g_contract, "virtual_bill_titles")()
        writer.writerow(
            ["MPRN", "Site Code", "Site Name", "Account", "From", "To"] + bill_titles
        )

        for month_start, month_finish in month_pairs:
            period_start = hh_max(start_date, month_start)
            period_finish = hh_min(finish_date, month_finish)

            for g_era in (
                sess.query(GEra)
                .distinct()
                .filter(
                    GEra.g_contract == g_contract,
                    GEra.start_date <= period_finish,
                    or_(GEra.finish_date == null(), GEra.finish_date >= period_start),
                )
            ):
                chunk_start = hh_max(g_era.start_date, period_start)
                chunk_finish = hh_min(g_era.finish_date, period_finish)

                data_source = GDataSource(
                    sess,
                    chunk_start,
                    chunk_finish,
                    forecast_dt,
                    g_era,
                    report_context,
                    None,
                )

                site = (
                    sess.query(Site)
                    .join(SiteGEra)
                    .filter(SiteGEra.g_era == g_era, SiteGEra.is_physical == true())
                    .one()
                )

                vals = [
                    data_source.mprn,
                    site.code,
                    site.name,
                    data_source.account,
                    hh_format(data_source.start_date),
                    hh_format(data_source.finish_date),
                ]

                contract_func(report_context, g_contract, "virtual_bill")(data_source)
                bill = data_source.bill
                for title in bill_titles:
                    if title in bill:
                        val = make_val(bill[title])
                        del bill[title]
                    else:
                        val = ""
                    vals.append(val)

                for k in sorted(bill.keys()):
                    vals.append(k)
                    vals.append(str(bill[k]))
                writer.writerow(vals)

    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    start_date = req_date("start")
    finish_date = req_date("finish")
    g_contract_id = req_int("g_contract_id")

    threading.Thread(
        target=content, args=(start_date, finish_date, g_contract_id, g.user)
    ).start()
    return chellow_redirect("/downloads", 303)



import csv
import os
import sys
import threading
import traceback

from dateutil.relativedelta import relativedelta

from flask import g, request

from sqlalchemy import or_
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import func, null

import chellow.dloads
from chellow.models import (
    Batch,
    Bill,
    Era,
    GeneratorType,
    MeasurementRequirement,
    MtcParticipant,
    ReadType,
    RegisterRead,
    Session,
    SiteEra,
    Supply,
)
from chellow.utils import (
    CHANNEL_TYPES,
    HH,
    csv_make_val,
    hh_format,
    hh_min,
    parse_mpan_core,
    req_date,
    req_int,
    req_str,
)
from chellow.views import chellow_redirect


def content(running_name, finished_name, date, supply_id, mpan_cores):
    sess = None
    try:
        sess = Session()
        f = open(running_name, mode="w", newline="")
        _process(sess, f, date, supply_id, mpan_cores)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        f.write(msg)
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def _process(sess, f, date, supply_id, mpan_cores):
    writer = csv.writer(f, lineterminator="\n")
    titles = (
        "Date",
        "Import MPAN Core",
        "Export MPAN Core",
        "Physical Site Id",
        "Physical Site Name",
        "Other Site Ids",
        "Other Site Names",
        "Supply Id",
        "Source",
        "Generator Type",
        "GSP Group",
        "DNO Name",
        "Voltage Level",
        "Is Substations",
        "Metering Type",
        "Mandatory HH",
        "PC",
        "MTC",
        "CoP",
        "Comms Type",
        "SSC Code",
        "SSC Description",
        "Energisation Status",
        "Number Of Registers",
        "MOP Contract",
        "Mop Account",
        "DC Contract",
        "DC Account",
        "Meter Serial Number",
        "Meter Installation Date",
        "Latest Normal Meter Read Date",
        "Latest Normal Meter Read Type",
        "Latest DC Bill Date",
        "Latest MOP Bill Date",
        "Supply Start Date",
        "Supply Finish Date",
        "Properties",
        "Import ACTIVE?",
        "Import REACTIVE_IMPORT?",
        "Import REACTIVE_EXPORT?",
        "Export ACTIVE?",
        "Export REACTIVE_IMPORT?",
        "Export REACTIVE_EXPORT?",
        "Import Agreed Supply Capacity (kVA)",
        "Import LLFC Code",
        "Import LLFC Description",
        "Import Supplier Contract",
        "Import Supplier Account",
        "Import Mandatory kW",
        "Latest Import Supplier Bill Date",
        "Export Agreed Supply Capacity (kVA)",
        "Export LLFC Code",
        "Export LLFC Description",
        "Export Supplier Contract",
        "Export Supplier Account",
        "Export Mandatory kW",
        "Latest Export Supplier Bill Date",
    )
    writer.writerow(titles)

    NORMAL_READ_TYPES = ("N", "C", "N3")
    year_start = date + HH - relativedelta(years=1)

    era_ids = (
        sess.query(Era.id)
        .filter(
            Era.start_date <= date,
            or_(Era.finish_date == null(), Era.finish_date >= date),
        )
        .order_by(Era.supply_id)
    )

    if supply_id is not None:
        supply = Supply.get_by_id(sess, supply_id)

        era_ids = era_ids.filter(Era.supply == supply)

    if mpan_cores is not None:
        era_ids = era_ids.filter(
            or_(Era.imp_mpan_core.in_(mpan_cores), Era.exp_mpan_core.in_(mpan_cores))
        )

    for (era_id,) in era_ids:
        era, supply, generator_type = (
            sess.query(Era, Supply, GeneratorType)
            .join(Supply, Era.supply_id == Supply.id)
            .outerjoin(GeneratorType, Supply.generator_type_id == GeneratorType.id)
            .filter(Era.id == era_id)
            .options(
                joinedload(Era.channels),
                joinedload(Era.cop),
                joinedload(Era.comm),
                joinedload(Era.dc_contract),
                joinedload(Era.exp_llfc),
                joinedload(Era.exp_supplier_contract),
                joinedload(Era.imp_llfc),
                joinedload(Era.imp_supplier_contract),
                joinedload(Era.mop_contract),
                joinedload(Era.mtc_participant).joinedload(MtcParticipant.mtc),
                joinedload(Era.mtc_participant).joinedload(MtcParticipant.meter_type),
                joinedload(Era.pc),
                joinedload(Era.site_eras).joinedload(SiteEra.site),
                joinedload(Era.ssc),
                joinedload(Era.energisation_status),
                joinedload(Era.supply).joinedload(Supply.source),
                joinedload(Era.supply).joinedload(Supply.gsp_group),
                joinedload(Era.supply).joinedload(Supply.dno),
            )
            .one()
        )

        site_codes = []
        site_names = []
        for site_era in era.site_eras:
            if site_era.is_physical:
                physical_site = site_era.site
            else:
                site = site_era.site
                site_codes.append(site.code)
                site_names.append(site.name)

        sup_eras = (
            sess.query(Era).filter(Era.supply == supply).order_by(Era.start_date).all()
        )
        supply_start_date = sup_eras[0].start_date
        supply_finish_date = sup_eras[-1].finish_date

        if era.imp_mpan_core is None:
            voltage_level_code = era.exp_llfc.voltage_level.code
            is_substation = era.exp_llfc.is_substation
        else:
            voltage_level_code = era.imp_llfc.voltage_level.code
            is_substation = era.imp_llfc.is_substation

        if generator_type is None:
            generator_type_str = ""
        else:
            generator_type_str = generator_type.code

        metering_type = era.meter_category

        if metering_type in ("nhh", "amr"):
            latest_prev_normal_read = (
                sess.query(RegisterRead)
                .join(Bill)
                .join(RegisterRead.previous_type)
                .filter(
                    ReadType.code.in_(NORMAL_READ_TYPES),
                    RegisterRead.previous_date <= date,
                    Bill.supply_id == supply.id,
                )
                .order_by(RegisterRead.previous_date.desc())
                .options(joinedload(RegisterRead.previous_type))
                .first()
            )

            latest_pres_normal_read = (
                sess.query(RegisterRead)
                .join(Bill)
                .join(RegisterRead.present_type)
                .filter(
                    ReadType.code.in_(NORMAL_READ_TYPES),
                    RegisterRead.present_date <= date,
                    Bill.supply == supply,
                )
                .order_by(RegisterRead.present_date.desc())
                .options(joinedload(RegisterRead.present_type))
                .first()
            )

            if latest_prev_normal_read is None and latest_pres_normal_read is None:
                latest_normal_read_date = None
                latest_normal_read_type = None
            elif (
                latest_pres_normal_read is not None and latest_prev_normal_read is None
            ):
                latest_normal_read_date = latest_pres_normal_read.present_date
                latest_normal_read_type = latest_pres_normal_read.present_type.code
            elif (
                latest_pres_normal_read is None and latest_prev_normal_read is not None
            ):
                latest_normal_read_date = latest_prev_normal_read.previous_date
                latest_normal_read_type = latest_prev_normal_read.previous_type.code
            elif (
                latest_pres_normal_read.present_date
                > latest_prev_normal_read.previous_date
            ):
                latest_normal_read_date = latest_pres_normal_read.present_date
                latest_normal_read_type = latest_pres_normal_read.present_type.code
            else:
                latest_normal_read_date = latest_prev_normal_read.previous_date
                latest_normal_read_type = latest_prev_normal_read.previous_type.code
            if latest_normal_read_date is not None:
                latest_normal_read_date = hh_format(latest_normal_read_date)

        else:
            latest_normal_read_date = metering_type
            latest_normal_read_type = None

        mop_contract = era.mop_contract
        mop_contract_name = mop_contract.name
        mop_account = era.mop_account
        latest_mop_bill_date = (
            sess.query(Bill.finish_date)
            .join(Batch)
            .filter(
                Bill.start_date <= date,
                Bill.supply == supply,
                Batch.contract == mop_contract,
            )
            .order_by(Bill.finish_date.desc())
            .first()
        )

        if latest_mop_bill_date is not None:
            latest_mop_bill_date = hh_format(latest_mop_bill_date[0])

        dc_contract = era.dc_contract
        dc_contract_name = dc_contract.name
        dc_account = era.dc_account
        latest_dc_bill_date = (
            sess.query(Bill.finish_date)
            .join(Batch)
            .filter(
                Bill.start_date <= date,
                Bill.supply == supply,
                Batch.contract == dc_contract,
            )
            .order_by(Bill.finish_date.desc())
            .first()
        )

        if latest_dc_bill_date is not None:
            latest_dc_bill_date = hh_format(latest_dc_bill_date[0])

        channel_values = []
        for imp_related in [True, False]:
            for channel_type in CHANNEL_TYPES:
                if era.find_channel(sess, imp_related, channel_type) is None:
                    channel_values.append("false")
                else:
                    channel_values.append("true")

        imp_avg_months = None
        exp_avg_months = None
        for is_import in [True, False]:
            if metering_type == "nhh":
                continue

            params = {
                "supply_id": supply.id,
                "year_start": year_start,
                "year_finish": date,
                "is_import": is_import,
            }
            month_mds = tuple(
                md[0] * 2
                for md in sess.execute(
                    """

select max(hh_datum.value) as md
from hh_datum join channel on (hh_datum.channel_id = channel.id)
    join era on (channel.era_id = era.id)
where era.supply_id = :supply_id and hh_datum.start_date >= :year_start
    and hh_datum.start_date <= :year_finish
    and channel.channel_type = 'ACTIVE'
    and channel.imp_related = :is_import
group by extract(month from (hh_datum.start_date at time zone 'utc'))
order by md desc
limit 3

""",
                    params=params,
                )
            )

            avg_months = sum(month_mds)
            if len(month_mds) > 0:
                avg_months /= len(month_mds)
                if is_import:
                    imp_avg_months = avg_months
                else:
                    exp_avg_months = avg_months

        if (imp_avg_months is not None and imp_avg_months > 100) or (
            exp_avg_months is not None and exp_avg_months > 100
        ):
            mandatory_hh = "yes"
        else:
            mandatory_hh = "no"

        imp_latest_supplier_bill_date = None
        exp_latest_supplier_bill_date = None
        for is_import in (True, False):
            for er in (
                sess.query(Era)
                .filter(Era.supply == era.supply, Era.start_date <= date)
                .order_by(Era.start_date.desc())
            ):
                if is_import:
                    if er.imp_mpan_core is None:
                        break
                    else:
                        supplier_contract = er.imp_supplier_contract
                else:
                    if er.exp_mpan_core is None:
                        break
                    else:
                        supplier_contract = er.exp_supplier_contract

                latest_bill_date = (
                    sess.query(Bill.finish_date)
                    .join(Batch)
                    .filter(
                        Bill.finish_date >= er.start_date,
                        Bill.finish_date <= hh_min(er.finish_date, date),
                        Bill.supply == supply,
                        Batch.contract == supplier_contract,
                    )
                    .order_by(Bill.finish_date.desc())
                    .first()
                )

                if latest_bill_date is not None:
                    latest_bill_date = hh_format(latest_bill_date[0])

                    if is_import:
                        imp_latest_supplier_bill_date = latest_bill_date
                    else:
                        exp_latest_supplier_bill_date = latest_bill_date
                    break

        meter_installation_date = (
            sess.query(func.min(Era.start_date))
            .filter(Era.supply == era.supply, Era.msn == era.msn)
            .one()[0]
        )

        ssc = era.ssc
        if ssc is None:
            ssc_code = ssc_description = num_registers = None
        else:
            ssc_code, ssc_description = ssc.code, ssc.description
            num_registers = (
                sess.query(MeasurementRequirement)
                .filter(MeasurementRequirement.ssc == ssc)
                .count()
            )

        vals = (
            [
                date,
                era.imp_mpan_core,
                era.exp_mpan_core,
                physical_site.code,
                physical_site.name,
                ", ".join(site_codes),
                ", ".join(site_names),
                supply.id,
                supply.source.code,
                generator_type_str,
                supply.gsp_group.code,
                supply.dno.dno_code,
                voltage_level_code,
                is_substation,
                metering_type,
                mandatory_hh,
                era.pc.code,
                era.mtc_participant.mtc.code,
                era.cop.code,
                era.comm.code,
                ssc_code,
                ssc_description,
                era.energisation_status.code,
                num_registers,
                mop_contract_name,
                mop_account,
                dc_contract_name,
                dc_account,
                era.msn,
                meter_installation_date,
                latest_normal_read_date,
                latest_normal_read_type,
                latest_dc_bill_date,
                latest_mop_bill_date,
                supply_start_date,
                supply_finish_date,
                era.properties,
            ]
            + channel_values
            + [
                era.imp_sc,
                None if era.imp_llfc is None else era.imp_llfc.code,
                None if era.imp_llfc is None else era.imp_llfc.description,
                None
                if era.imp_supplier_contract is None
                else era.imp_supplier_contract.name,
                era.imp_supplier_account,
                imp_avg_months,
                imp_latest_supplier_bill_date,
            ]
            + [
                era.exp_sc,
                None if era.exp_llfc is None else era.exp_llfc.code,
                None if era.exp_llfc is None else era.exp_llfc.description,
                None
                if era.exp_supplier_contract is None
                else era.exp_supplier_contract.name,
                era.exp_supplier_account,
                exp_avg_months,
                exp_latest_supplier_bill_date,
            ]
        )
        writer.writerow([csv_make_val(v) for v in vals])

        # Avoid a long-running transaction
        sess.rollback()


def do_get(session):
    user = g.user
    date = req_date("date")
    if "supply_id" in request.values:
        supply_id = req_int("supply_id")
    else:
        supply_id = None

    if "mpan_cores" in request.values:
        mpan_cores_str = req_str("mpan_cores")
        mpan_cores = mpan_cores_str.splitlines()
        if len(mpan_cores) == 0:
            mpan_cores = None
        else:
            for i in range(len(mpan_cores)):
                mpan_cores[i] = parse_mpan_core(mpan_cores[i])
    else:
        mpan_cores = None

    running_name, finished_name = chellow.dloads.make_names(
        "supplies_snapshot.csv", user
    )

    args = (running_name, finished_name, date, supply_id, mpan_cores)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import csv
from io import StringIO

import chellow.reports.report_g_supplies_snapshot
from chellow.models import (
    GContract,
    GDn,
    GReadingFrequency,
    GUnit,
    MarketRole,
    Participant,
    Site,
    insert_g_reading_frequencies,
    insert_g_units,
)
from chellow.utils import utc_datetime


def test_supply(mocker, sess, client):
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    msn = "hgeu8rhg"
    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2010, 1, 1),
        None,
        msn,
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )
    sess.commit()

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch(
        "chellow.reports.report_g_supplies_snapshot.open", return_value=mock_file
    )
    mocker.patch(
        "chellow.reports.report_g_supplies_snapshot.chellow.dloads.make_names",
        return_value=("a", "b"),
    )
    mocker.patch("chellow.reports.report_g_supplies_snapshot.os.rename")

    user = mocker.Mock()
    g_supply_id = g_supply.id
    date = utc_datetime(2020, 9, 1)

    chellow.reports.report_g_supplies_snapshot.content(date, g_supply_id, user)

    mock_file.seek(0)
    sheet = csv.reader(mock_file)
    table = list(sheet)

    expected = [
        [
            "Date",
            "Physical Site Id",
            "Physical Site Name",
            "Other Site Ids",
            "Other Site Names",
            "MPRN",
            "Exit Zone",
            "Meter Serial Number",
            "Correction Factor",
            "Unit",
            "Contract",
            "Account",
            "Supply Start",
            "Supply Finish",
        ],
        [
            "2020-09-01 01:00",
            "22488",
            "Water Works",
            "",
            "",
            "87614362",
            "EA1",
            "hgeu8rhg",
            "1",
            "M3",
            "Fusion 2020",
            "d7gthekrg",
            "2010-01-01 00:00",
            "",
        ],
    ]

    assert expected == table



import atexit
import collections
import csv
import threading
import traceback
from decimal import Decimal

from dateutil.relativedelta import relativedelta

import requests

from werkzeug.exceptions import BadRequest

from zish import loads

from chellow.models import Contract, RateScript, Session, get_non_core_contract_id
from chellow.utils import (
    ct_datetime_parse,
    hh_format,
    to_utc,
    u_months_u,
    utc_datetime_now,
)


ELEXON_PORTAL_SCRIPTING_KEY_KEY = "elexonportal_scripting_key"


def hh(data_source):
    try:
        cache = data_source.caches["rcrc"]
    except KeyError:
        cache = data_source.caches["rcrc"] = {}

    for hh in data_source.hh_data:
        try:
            hh["rcrc-rate"] = rcrc = cache[hh["start-date"]]
        except KeyError:
            h_start = hh["start-date"]
            db_id = get_non_core_contract_id("rcrc")
            rates = data_source.hh_rate(db_id, h_start)["rates"]
            try:
                hh["rcrc-rate"] = rcrc = cache[h_start] = (
                    float(rates[key_format(h_start)]) / 1000
                )
            except KeyError:
                try:
                    dt = h_start - relativedelta(days=3)
                    hh["rcrc-rate"] = rcrc = cache[h_start] = (
                        float(rates[key_format(dt)]) / 1000
                    )
                except KeyError:
                    raise BadRequest(
                        f"For the RCRC rate script at {hh_format(dt)} the rate cannot "
                        f"be found."
                    )

        hh["rcrc-kwh"] = hh["nbp-kwh"]
        hh["rcrc-gbp"] = hh["nbp-kwh"] * rcrc


rcrc_importer = None


def key_format(dt):
    return dt.strftime("%d %H:%M Z")


class RcrcImporter(threading.Thread):
    def __init__(self):
        super(RcrcImporter, self).__init__(name="RCRC Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=100)
        self.stopped = threading.Event()
        self.going = threading.Event()

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                sess = None
                try:
                    sess = Session()
                    _process(self.log, sess)
                except BaseException:
                    self.log("Outer problem " + traceback.format_exc())
                    sess.rollback()
                finally:
                    self.lock.release()
                    self.log("Finished checking RCRC rates.")
                    if sess is not None:
                        sess.close()

            self.going.wait(30 * 60)
            self.going.clear()


def _find_month(lines, month_start, month_finish):
    parser = csv.reader(lines, delimiter=",", quotechar='"')
    next(parser)
    next(parser)
    month_rcrcs = {}
    for values in parser:
        hh_date = to_utc(ct_datetime_parse(values[0], "%d/%m/%Y"))
        hh_date += relativedelta(minutes=30 * int(values[2]))
        if month_start <= hh_date <= month_finish:
            month_rcrcs[key_format(hh_date)] = Decimal(values[3])
    return month_rcrcs


def _process(log_f, sess):
    log_f("Starting to check RCRCs.")
    contract = Contract.get_non_core_by_name(sess, "rcrc")
    latest_rs = (
        sess.query(RateScript)
        .filter(RateScript.contract_id == contract.id)
        .order_by(RateScript.start_date.desc())
        .first()
    )
    latest_rs_id = latest_rs.id
    latest_rs_start = latest_rs.start_date

    months = list(
        u_months_u(
            start_year=latest_rs_start.year, start_month=latest_rs_start.month, months=2
        )
    )
    month_start, month_finish = months[1]
    now = utc_datetime_now()
    if now > month_finish:
        config = Contract.get_non_core_by_name(sess, "configuration")
        props = config.make_properties()

        scripting_key = props.get(ELEXON_PORTAL_SCRIPTING_KEY_KEY)
        if scripting_key is None:
            raise BadRequest(
                f"The property {ELEXON_PORTAL_SCRIPTING_KEY_KEY} cannot be found in "
                f"the configuration properties."
            )

        contract_props = contract.make_properties()
        url_str = f"{contract_props['url']}file/download/RCRC_FILE?key={scripting_key}"
        log_f(
            f"Downloading {url_str} to see if data is available from "
            f"{hh_format(month_start)} to {hh_format(month_finish)}."
        )

        sess.rollback()  # Avoid long-running transaction
        r = requests.get(url_str, timeout=60)
        month_rcrcs = _find_month(
            (x.decode() for x in r.iter_lines()), month_start, month_finish
        )
        if key_format(month_finish) in month_rcrcs:
            log_f("The whole month's data is there.")
            script = {"rates": month_rcrcs}
            contract = Contract.get_non_core_by_name(sess, "rcrc")
            rs = RateScript.get_by_id(sess, latest_rs_id)
            contract.update_rate_script(
                sess, rs, rs.start_date, month_finish, loads(rs.script)
            )
            contract.insert_rate_script(sess, month_start, script)
            sess.commit()
            log_f(f"Added a new rate script starting at {hh_format(month_start)}.")
        else:
            msg = "There isn't a whole month there yet."
            if len(month_rcrcs) > 0:
                msg += f" The last date is {sorted(month_rcrcs.keys())[-1]}"
            log_f(msg)


def get_importer():
    return rcrc_importer


def startup():
    global rcrc_importer
    rcrc_importer = RcrcImporter()
    rcrc_importer.start()


@atexit.register
def shutdown():
    if rcrc_importer is not None:
        rcrc_importer.stop()
